import abc
from collections import deque

from mimircache.cache.abstractCache import cache


class abstractLFU(cache):
    def __init__(self, cache_size=1000):
        super().__init__(cache_size)
        self.cacheDict = dict()  # key -> freq
        self.largest_freq = -1
        self.largest_freq_element = None
        self.least_freq = 1
        self.least_freq_elements_list = deque()
        self.least_freq_elements_set = set()

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.cacheDict:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now its frequency
        :param element:
        :return: original rank
        """
        self.cacheDict[element] += 1
        # if self.cache_dict[element]>self.largest_freq:
        #     self.largest_freq = self.cache_dict[element]
        #     self.largest_freq_element = element
        if element in self.least_freq_elements_set:
            if len(self.least_freq_elements_set) > 1:
                # more than one element, so just remove this element
                self.least_freq_elements_set.remove(element)
                self.least_freq_elements_list.remove(element)
            else:
                # this is the only element, keep it in the set/list, add more with same freq
                for key, value in self.cacheDict.items():
                    if value == self.cacheDict[element] and key != element:
                        self.least_freq_elements_set.add(key)
                        self.least_freq_elements_list.append(key)

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        # print("***%s***"%element)
        if len(self.cacheDict) >= self.cache_size:
            self._evictOneElement()

        self.cacheDict[element] = 1
        # if self.cache_dict[element]>self.largest_freq:
        #     self.largest_freq = self.cache_dict[element]
        #     self.largest_freq_element = element
        if self.least_freq == 1:
            # this one needs to be added
            self.least_freq_elements_list.append(element)
            self.least_freq_elements_set.add(element)
        elif self.least_freq > 1:
            self.least_freq = 1
            self.least_freq_elements_list.clear()
            self.least_freq_elements_set.clear()

            self.least_freq_elements_list.append(element)
            self.least_freq_elements_set.add(element)

    @abc.abstractmethod
    def find_evict_key(self):
        pass

    def _printCacheLine(self):
        for key, value in self.cacheDict.items():
            print("{}: {}".format(key, value))

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        evict_key = self.find_evict_key()
        self.least_freq_elements_set.remove(evict_key)
        del self.cacheDict[evict_key]
        while len(self.least_freq_elements_list) == 0:
            if len(self.cacheDict) == 0:
                break
            # after remove, there is no element in the least frequent set
            self.least_freq += 1
            for key, value in self.cacheDict.items():
                if value == self.least_freq:
                    self.least_freq_elements_list.append(key)
                    self.least_freq_elements_set.add(key)

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: -1 if not in cache, otherwise old rank
        """
        # print(element, end=': \t')
        if self.checkElement(element):
            self._updateElement(element)
            return True
        else:
            self._insertElement(element)
            return False

    def __repr__(self):
        return "LFU, given size: {}, current size: {}".format(self.cache_size, len(self.cacheDict))

    def __str__(self):
        return self.__repr__()



import os
from ctypes import *
import logging

from mimircache.cacheReader.abstractReader import cacheReaderAbstract
import mimircache.c_cacheReader as c_cacheReader


class vscsiCacheReader(cacheReaderAbstract):
    def __init__(self, file_loc):
        super().__init__(file_loc)
        self.cReader = c_cacheReader.setup_reader(file_loc, 'v')

        self.buffer_size = 0  # number of trace lines in the buffer
        self.buffer_pointer = 0  # point to the element in the buffer
        self.read_in_num = 0

        self.vscsiC = CDLL(os.path.join(os.path.dirname(os.path.abspath(__file__)), self.get_lib_name()))
        self.vscsiC.argTypes = [c_char_p, c_void_p, POINTER(c_int), POINTER(c_int), POINTER(c_int)]
        self.c_mem = c_void_p()
        self.c_ver = c_int(-1)
        self.c_delta = c_int(-1)
        self.c_num_of_rec = c_int(-1)

        self.c_read_size = c_int()
        self.c_long_ts = (c_long * 10000)()
        self.c_int_len = (c_int * 10000)()
        self.c_long_lbn = (c_long * 10000)()
        self.c_int_cmd = (c_int * 10000)()

        try:
            # python 2
            filename = c_char_p(self.file_loc)
        except:
            # python 3
            filename = c_char_p(self.file_loc.encode())

        self.vscsiC.setup(filename, byref(self.c_mem), byref(self.c_ver), byref(self.c_delta), byref(self.c_num_of_rec))
        # self.vscsiC.read_trace.resType = c_long
        self.vscsiC.read_trace2.resType = c_int
        self.mem_original = c_void_p(self.c_mem.value)
        self._read()
        self.get_num_of_total_requests()


    def get_lib_name(self):
        for name in os.listdir(os.path.dirname(os.path.abspath(__file__))):
            if 'libvscsi' in name and '.py' not in name:
                return name


    def get_first_line(self):
        self.reset()
        return next(self.lines())

    def get_last_line(self):
        self.c_mem = c_void_p(self.mem_original.value + self.c_delta.value * (self.c_num_of_rec.value - 1))
        self.vscsiC.read_trace2(byref(self.c_mem), byref(self.c_ver), c_int(1), self.c_long_ts, self.c_int_len,
                                self.c_long_lbn, self.c_int_cmd)
        return self.c_long_ts[0], self.c_int_cmd[0], self.c_int_len[0], self.c_long_lbn[0]


    def reset(self):
        logging.debug("reset")
        self.buffer_pointer = 0
        self.read_in_num = 0
        self.counter = 0
        self.c_read_size.value = 0
        self.c_mem = c_void_p(self.mem_original.value)
        c_cacheReader.reset_reader(self.cReader)

    def read_one_element(self):
        if self.buffer_pointer < self.c_read_size.value:
            self.buffer_pointer += 1
            return self.c_long_lbn[self.buffer_pointer - 1]

        elif self.read_in_num < self.c_num_of_rec.value:
            self._read()
            self.buffer_pointer = 1
            return self.c_long_lbn[self.buffer_pointer - 1]
        else:
            return None

    def get_num_of_total_requests(self):
        return self.c_num_of_rec.value


    def lines(self):
        # ts, cmd, size, lbn
        """
        return detailed information in vscsi reader
        :return:
        """
        self.reset()
        while self.read_in_num < self.c_num_of_rec.value or self.buffer_pointer < self.c_read_size.value:
            if self.buffer_pointer < self.c_read_size.value:
                self.buffer_pointer += 1
                yield (self.c_long_ts[self.buffer_pointer - 1], self.c_int_cmd[self.buffer_pointer - 1],
                       self.c_int_len[self.buffer_pointer - 1], self.c_long_lbn[self.buffer_pointer - 1])

            elif self.read_in_num < self.c_num_of_rec.value:
                self._read()
                self.buffer_pointer = 1
                yield (self.c_long_ts[self.buffer_pointer - 1], self.c_int_cmd[self.buffer_pointer - 1],
                       self.c_int_len[self.buffer_pointer - 1], self.c_long_lbn[self.buffer_pointer - 1])

    def __next__(self):  # Python 3
        super().__next__()
        if self.buffer_pointer < self.c_read_size.value:
            self.buffer_pointer += 1
            return self.c_long_lbn[self.buffer_pointer - 1]

        elif self.read_in_num < self.c_num_of_rec.value:
            self._read()
            self.buffer_pointer = 1
            return self.c_long_lbn[self.buffer_pointer - 1]
        else:
            raise StopIteration

    def _read(self):
        if self.c_num_of_rec.value - self.read_in_num >= 10000:
            self.c_read_size = c_int(10000)
        else:
            self.c_read_size = c_int(self.c_num_of_rec.value - self.read_in_num)
        self.read_in_num += self.c_read_size.value
        self.vscsiC.read_trace2(byref(self.c_mem), byref(self.c_ver), self.c_read_size, self.c_long_ts, self.c_int_len,
                                self.c_long_lbn, self.c_int_cmd)

    # TODO: needs cleaning, otherwise it might take huge RAM!!!!!!!!!


    def __repr__(self):
        return "vscsi cache reader, %s" % super().__repr__()


if __name__ == "__main__":
    reader = vscsiCacheReader('../data/trace_CloudPhysics_bin')

    # # usage one: for reading all elements
    num = 0
    print(reader.c_num_of_rec)
    for line in reader.lines():
        if num == 0:
            prev = line[0]
            num += 1
            continue
        print(line[0] - prev)
        prev = line[0]

        num += 1


        # print("{}: {}".format(num, i))
        # print(num)
        # print(reader)
        # print(reader.get_first_line())
        # print(reader.get_last_line())

    #
    # # usage two: best for reading one element each time
    # s = reader.read_one_element()
    # # s2 = next(reader)
    # while (s):
    #     print(s)
    #     s = reader.read_one_element()
    #     # s2 = next(reader)

    # test3: read first 10 elements twice
    # for i in range(10):
    #     print(reader.read_one_element())
    # print("after reset")
    # reader.reset()
    # for i in range(10):
    #     print(reader.read_one_element())



from pylab import *
from numpy import outer

rc('text', usetex=False)
a = outer(arange(0, 1, 0.01), ones(10))
figure(figsize=(10, 5))
subplots_adjust(top=0.8, bottom=0.05, left=0.01, right=0.99)
maps = [m for m in cm.datad if not m.endswith("_r")]
maps.sort()
l = len(maps) + 1
for i, m in enumerate(maps):
    subplot(1, l, i + 1)
    axis("off")
    imshow(a, aspect='auto', cmap=get_cmap(m), origin="lower")
    title(m, rotation=90, fontsize=10)
savefig("colormaps.png", dpi=100, facecolor='gray')



import abc
import os
from multiprocessing import Lock
import mimircache.c_cacheReader as c_cacheReader


class cacheReaderAbstract(metaclass=abc.ABCMeta):
    __metaclass__ = abc.ABCMeta

    @abc.abstractclassmethod
    def __init__(self, file_loc):
        self.file_loc = file_loc
        assert (os.path.exists(file_loc)), "data file does not exist"
        # self.trace_file = open(file_loc, 'r')
        self.counter = 0
        self.trace_file = None
        self.cReader = None
        self.num_of_line = -1
        self.lock = Lock()

    def reset(self):
        """
        reset the read location back to beginning
        :return:
        """
        self.counter = 0
        self.trace_file.seek(0, 0)
        c_cacheReader.reset_reader(self.cReader)

    def get_num_of_total_requests(self):
        if self.num_of_line != -1:
            return self.num_of_line

        if self.cReader:
            self.num_of_line = c_cacheReader.get_num_of_lines(self.cReader)
        else:
            while self.read_one_element():
                # use the one above, not output progress
                self.num_of_line += 1
            self.reset()

        return self.num_of_line

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    @abc.abstractclassmethod
    def read_one_element(self):
        pass
        # self.counter += 1
        # if (self.counter % 1000 == 0):
        #     print('read in ' + str(self.counter) + ' records')
        # raise NotImplementedError

    def close(self):
        try:
            # pass
            self.trace_file.close()
            if self.cReader:
                # pass
                c_cacheReader.close_reader(self.cReader)
        except:
            pass

    @abc.abstractclassmethod
    def __next__(self):  # Python 3
        self.counter += 1
        # if (self.counter % 100000 == 0):
        #     print('read in ' + str(self.counter) + ' records')
        # raise NotImplementedError

    # @atexit.register
    def __del__(self):
        self.close()



import traceback
import abc
import math
from os import path

# deal with headless situation
# if Constants.headless:
import matplotlib
# matplotlib.use('Agg')

import matplotlib.pyplot as plt

from mimircache.cache.LRU import LRU
from mimircache.profiler.abstract.abstractProfiler import profilerAbstract

DEBUG = False


class abstractLRUProfiler(profilerAbstract):
    __metaclass__ = abc.ABCMeta

    @abc.abstractclassmethod
    def __init__(self, cache_class, cache_size, bin_size, reader):
        super().__init__(cache_class, cache_size, reader)
        self.bin_size = bin_size
        if cache_size != -1:
            self.num_of_cache = self.num_of_blocks = math.ceil(cache_size / bin_size) + 1
            self.HRC = [0] * self.num_of_blocks
            self.MRC = [0] * self.num_of_blocks
            self.HRC_bin = [0] * self.num_of_blocks
        assert self.cache_class == LRU
        self.cache = self.cache_class(self.cache_size)
        self.calculated = False

    @abc.abstractclassmethod
    def addOneTraceElement(self, element):
        super().addOneTraceElement(element)
        self.calculated = False
        return

    def calculate(self):
        if DEBUG:
            print(self.HRC_bin)
        self.HRC[0] = self.HRC_bin[0] / self.num_of_trace_elements * 100
        for i in range(1, len(self.HRC)):
            self.HRC[i] = self.HRC[i - 1] + self.HRC_bin[i] / self.num_of_trace_elements * 100
        if DEBUG:
            print(self.HRC)
        self.calculated = True
        for i in range(len(self.HRC)):
            self.MRC[i] = 100 - self.HRC[i]

    def printMRC(self):
        if not self.calculated:
            self.calculate()
        for x in self.MRC:
            print(x)

    def printHRC(self):
        if not self.calculated:
            self.calculate()
        for x in self.HRC:
            print(x)

    def returnMRC(self):
        if not self.calculated:
            self.calculate()
        return self.MRC

    def returnHRC(self):
        if not self.calculated:
            self.calculate()
        return self.HRC

    def plotMRC(self, autosize=False, autosize_threshold=0.01):
        if not self.calculated:
            self.calculate()
        try:
            # p_MRC = figure_MRC.add_subplot(1,1,1)

            # change the x-axis range according to threshhold
            if autosize:
                i = 0
                for i in range(len(self.MRC) - 1, 2, -1):
                    if (self.MRC[i - 1] - self.MRC[i]) / self.MRC[i] > autosize_threshold:
                        break
                num_of_blocks = i

            else:
                num_of_blocks = self.num_of_blocks
            if DEBUG:
                print("number of blocks: {}".format(num_of_blocks))
                print("total size: {}".format(self.bin_size * num_of_blocks))

            plt.plot(range(0, self.bin_size * num_of_blocks, self.bin_size), self.MRC[:num_of_blocks])
            plt.xlabel("cache Size")
            plt.ylabel("Miss Rate/%")
            plt.title('Miss Rate Curve', fontsize=18, color='black')
            plt.show()
            plt.savefig("figure_temp_MRC.pdf")
        except Exception as e:
            plt.savefig("figure_temp_MRC.pdf")
            print("the plotting function is not wrong, is this a headless server?")
            print(e)
            traceback.print_exc()

    def plotHRC(self, autosize=False, autosize_threshold=0.001):
        if not self.calculated:
            self.calculate()
        try:
            figure_HRC = plt.figure(2)

            # change the x-axis range according to threshhold
            if autosize:
                for i in range(len(self.HRC) - 1, 2, -1):
                    if (self.HRC[i] - self.HRC[i - 1]) / self.HRC[i] > autosize_threshold:
                        break
                num_of_blocks = i

            else:
                num_of_blocks = self.num_of_blocks
            if DEBUG:
                print("number of blocks: {}".format(num_of_blocks))
                print("total size: {}".format(self.bin_size * num_of_blocks))

            line = plt.plot(range(0, self.bin_size * num_of_blocks, self.bin_size), self.HRC[:num_of_blocks])
            plt.xlabel("cache Size")
            plt.ylabel("Hit Rate/%")
            plt.title('Hit Rate Curve', fontsize=18, color='black')
            plt.show()
            plt.savefig("figure_temp_HRC.pdf")
        except Exception as e:
            plt.savefig("figure_temp_HRC.pdf")
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def outputMRC(self, file_loc=""):
        """ save the MRC list to the given location
        :param file_loc: can be a fileName or foldername
        :return:
        """
        if not self.calculated:
            self.calculate()

        if path.isfile(file_loc):
            with open(file_loc, 'w') as file:
                for x in self.MRC:
                    file.write(str(x) + '\n')
        elif path.isdir(file_loc):
            with open(file_loc + '/' + str(self.cache_class.__name__) + '.MRC', 'w') as file:
                for x in self.MRC:
                    file.write(str(x) + '\n')
        return True

    def outputHRC(self, file_loc=""):
        """ save the MRC list to the given location
        :param file_loc: can be a fileName or foldername
        :return:
        """
        if not self.calculated:
            self.calculate()

        if path.isfile(file_loc):
            with open(file_loc, 'w') as file:
                for x in self.HRC:
                    file.write(str(x) + '\n')
        elif path.isdir(file_loc):
            with open(file_loc + '/' + str(self.cache_class.__name__) + '.HRC', 'w') as file:
                for x in self.HRC:
                    file.write(str(x) + '\n')
        return True



""" this module is used for all other cache replacement algorithms including LRU,
    but for LRU, we can use reuse distance for more efficient profiling, see basicLRUProfiler

"""
import math
import os

# deal with headless situation
# import matplotlib
# matplotlib.use('Agg')

from multiprocessing import Process, Pipe, Array
from os import path
from mimircache.cache.ARC import ARC
from mimircache.cache.clock import clock
from mimircache.cache.FIFO import FIFO
from mimircache.cache.LFU_LRU__NEED_OPTIMIZATION import LFU_LRU
from mimircache.cache.LFU_MRU import LFU_MRU
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.MRU import MRU
from mimircache.cache.Random import Random
from mimircache.cache.SLRU import SLRU
from mimircache.cache.S4LRU import S4LRU

from mimircache.cacheReader.plainReader import plainCacheReader

import matplotlib.pyplot as plt

from mimircache.profiler.abstract.abstractProfiler import profilerAbstract

debug = True


class generalProfiler(profilerAbstract):
    def __init__(self, cache_class, cache_params, bin_size, reader, num_of_process):
        if isinstance(cache_params, tuple):
            super(generalProfiler, self).__init__(cache_class, cache_params[0], reader)
            self.cache_size = cache_params[0]
            self.cache_params = cache_params[1:]

        elif isinstance(cache_params, int):
            super(generalProfiler, self).__init__(cache_class, cache_params, reader)
            self.cache_size = cache_params
            self.cache_params = ()

        else:
            raise RuntimeError("cache parameter not corrected, given: {}, expect either int as cache size or \
            tuple for full argument".format(cache_params))

        self.bin_size = bin_size
        self.num_of_process = num_of_process

        if self.cache_size != -1:
            if int(self.cache_size / bin_size) * bin_size == self.cache_size:
                self.num_of_cache = self.num_of_blocks = int(self.cache_size / bin_size)
            else:
                self.num_of_cache = self.num_of_blocks = math.ceil(self.cache_size / bin_size) + 1
            self.HRC = [0] * self.num_of_blocks
            self.MRC = [0] * self.num_of_blocks

        self.cache = None
        self.cache_list = None

        self.process_list = []
        self.cache_distribution = [[] for _ in range(self.num_of_process)]

        # shared memory for storing MRC count
        self.MRC_array = Array('i', range(self.num_of_blocks))
        for i in range(len(self.MRC_array)):
            self.MRC_array[i] = 0

        # dispatch different cache size to different processes
        for i in range(self.num_of_blocks):
            self.cache_distribution[i % self.num_of_process].append((i + 1) * bin_size)
            # print(self.cache_distribution)

            # build pipes for communication between main process and children process
            # the pipe mainly sends element from main process to children

        self.pipe_list = []

        for i in range(self.num_of_process):
            self.pipe_list.append(Pipe())
            p = Process(target=self._addOneTraceElementSingleProcess,
                        args=(self.num_of_process, i, self.cache_class, self.cache_distribution[i],
                              self.cache_params, self.pipe_list[i][1], self.MRC_array))
            self.process_list.append(p)
            p.start()

        self.calculated = False
        self._auto_output_filename = str(os.path.basename(self.reader.file_loc).split('.')[
                                             0]) + '_' + self.cache_class.__name__ + '_' + str(self.cache_size)

    def addOneTraceElement(self, element):
        super().addOneTraceElement(element)

        for i in range(len(self.pipe_list)):
            # print("send out: " + element)
            self.pipe_list[i][0].send(element)

        return

    def _addOneTraceElementSingleProcess(self, num_of_process, process_num, cache_class, cache_size_list,
                                         cache_args, pipe, MRC_array):
        '''

        :param num_of_process:
        :param process_num:
        :param cache_class:
        :param cache_size_list: a list of different cache size dispached to this process
        :param cache_args: the extra argument (besides size) for instantiate a cache class
        :param pipe:            for receiving cache record from main process
        :param MRC_array:       storing MRC count for all cache sizes
        :return:
        '''
        cache_list = []
        for i in range(len(cache_size_list)):
            if len(cache_args) > 0:
                cache_list.append(cache_class(cache_size_list[i], *cache_args))
            else:
                cache_list.append(cache_class(cache_size_list[i]))

        element = pipe.recv()

        # TODO this part should be changed
        while element != 'END_1a1a11a_ENDMARKER':
            for i in range(len(cache_list)):
                # print("i = %d"%i)
                # cache_list[i].printCacheLine()
                # print('')
                if not cache_list[i].addElement(element):
                    MRC_array[i * num_of_process + process_num] += 1
            element = pipe.recv()
            # print(element)
            # print(cache_list)

    def printMRC(self):
        if not self.calculated:
            self.calculate()
        for x in self.MRC:
            print(x)

    def printHRC(self):
        if not self.calculated:
            self.calculate()
        for x in self.HRC:
            print(x)

    def returnMRC(self):
        if not self.calculated:
            self.calculate()
        return self.MRC

    def returnHRC(self):
        if not self.calculated:
            self.calculate()
        return self.HRC

    def plotMRC(self):
        if not self.calculated:
            self.calculate()
        try:
            figure_MRC = plt.figure(1)
            line = plt.plot(range(self.bin_size, self.bin_size * (self.num_of_blocks + 1), self.bin_size), self.MRC)
            plt.xlabel("cache Size")
            plt.ylabel("Miss Rate/%")
            plt.title('Miss Rate Curve', fontsize=18, color='black')
            plt.show()
            plt.savefig(
                str(os.path.basename(self.reader.file_loc).split('.')[0]) + '_' + self.cache_class.__name__ + '_' + str(
                    self.cache_size) + '_MRC')
        except Exception as e:
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def plotHRC(self):
        if not self.calculated:
            self.calculate()
        try:
            figure_HRC = plt.figure(2)
            line = plt.plot(range(self.bin_size, self.bin_size * (self.num_of_blocks + 1), self.bin_size), self.HRC)
            plt.xlabel("cache Size")
            plt.ylabel("Hit Rate/%")
            plt.title('Hit Rate Curve', fontsize=18, color='black')
            plt.show()
            plt.savefig(
                str(os.path.basename(self.reader.file_loc).split('.')[0]) + '_' + self.cache_class.__name__ + '_' + str(
                    self.cache_size) + '_HRC')
            print("plot saved")
        except Exception as e:
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def outputMRC(self, file_loc="./"):
        """ save the MRC list to the given location
        :param file_loc: can be a fileName or foldername
        :return:
        """
        if not self.calculated:
            self.calculate()

        if path.isfile(file_loc):
            with open(file_loc, 'w') as file:
                for x in self.MRC:
                    file.write(str(x) + '\n')
        elif path.isdir(file_loc):
            with open(file_loc + '/' + str(self.cache_class.__name__) + '.MRC', 'w') as file:
                for x in self.MRC:
                    file.write(str(x) + '\n')
        return True

    def outputHRC(self, file_loc="./"):
        """ save the HRC list to the given location
        :param file_loc: can be a fileName or foldername
        :return:
        """
        if not self.calculated:
            self.calculate()

        if path.isfile(file_loc):
            with open(file_loc, 'w') as file:
                for x in self.HRC:
                    file.write(str(x) + '\n')
        elif path.isdir(file_loc):
            with open(self._auto_output_filename + '.HRC', 'w') as file:
                for x in self.HRC:
                    file.write(str(x) + '\n')
        return True

    def calculate(self):
        self.calculated = True
        for i in range(len(self.pipe_list)):
            self.pipe_list[i][0].send("END_1a1a11a_ENDMARKER")
            self.pipe_list[i][0].close()
        for i in range(len(self.process_list)):
            self.process_list[i].join()

        for i in range(len(self.MRC)):
            # print(self.MRC_array[i])
            self.MRC[i] = self.MRC_array[i] * 100 / self.num_of_trace_elements
        for i in range(len(self.HRC)):
            self.HRC[i] = 100 - self.MRC[i]

    def run(self):
        super().run()
        self.reader.reset()
        for i in self.reader:
            self.addOneTraceElement(i)
        # p.printMRC()
        self.outputHRC()
        self.plotHRC()


if __name__ == "__main__":
    import time
    import cProfile

    t1 = time.time()
    # r = plainCacheReader('../../data/test')
    r = plainCacheReader('../../data/parda.trace')

    # p = generalProfiler(LRU, 6000, 20, r, 48)
    # p = generalProfiler(ARC, (10, 0.5), 10, r, 1)
    p = generalProfiler(LFU_LRU, 800, 80, r, 4)
    # p = generalProfiler(ARC, 5, 5, r, 1)


    import pstats, io

    pr = cProfile.Profile()
    pr.enable()

    p.run()

    pr.disable()
    s = io.StringIO()
    sortby = 'cumulative'
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())

    t2 = time.time()
    print("TIME: %f" % (t2 - t1))

    # for i in r:
    #     # print(i)
    #     p.addOneTraceElement(i)
    # # p.printMRC()
    # p.plotHRC()



from mimircache.cache.abstractCache import cache
from mimircache.utils.LinkedList import LinkedList


class LRU(cache):
    def __init__(self, cache_size=1000):
        super().__init__(cache_size)
        self.cacheLinkedList = LinkedList()
        self.cacheDict = dict()  # store the reuse dist, thus dict: content->reuse dist, rank begins from 1

    def checkElement(self, content):
        """
        :param content: the content for search
        :return: whether the given element is in the cache
        """
        if content in self.cacheDict:
            return True
        else:
            return False

    def getReuseDist(self, content):
        """

        :param content: the content of element, which is also the key in cache_dict
        :return: rank if in the cache, otherwise -1
        """
        return self.cacheDict.get(content, -1)

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: original rank
        """
        rank = self.getReuseDist(element)
        # if rank >= 10:
        #     print("WWWWWWWWHHHHHHHHAAAAAAATTTTTTT")
        #     print(len(self.cache_dict))
        #     print(element)
        #     print(self.cache_dict)
        #     self.printLinkedList()
        # even if reuse distance is 0, it still needs at least one cache line
        self.cacheDict[element] = 1
        node = None
        counter = 0
        for node in self.cacheLinkedList:
            # print(node.content, end='\t')
            if node is None:
                print("**************stop**************")
            if node.content == element:
                break
            self.cacheDict[node.content] += 1
            counter += 1
        if counter != rank - 1:
            print("error")
            print(str(rank) + ":\t" + str(counter))
        self.cacheLinkedList.moveNodeToHead(node)
        return rank

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        self.cacheLinkedList.insertAtHead(element)
        self.cacheDict[element] = 0
        # above is set as 0 because it will increment by 1 in the following loop
        for i in self.cacheDict.keys():
            self.cacheDict[i] += 1
        if self.cacheLinkedList.size > self.cache_size:
            self._evictOneElement()

    def _printCacheLine(self):
        for i in self.cacheLinkedList:
            try:
                print(i.content, end='\t')
            except:
                print(i.content)

        print(' ')

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        content = self.cacheLinkedList.removeFromTail()
        del self.cacheDict[content]

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: -1 if not in cache, otherwise old rank
        """
        # print(element, end=': \t')
        if self.checkElement(element):
            rank = self._updateElement(element)
            # print(self.cache_dict)
            # self.printLinkedList()
            return rank
        else:
            self._insertElement(element)
            # print(self.cache_dict)
            # self.printLinkedList()
            return -1

    def __repr__(self):
        return "LRU, size: {}, {}".format(self.cache_size, super().__repr__())



import abc


class profilerAbstract(metaclass=abc.ABCMeta):
    __metaclass__ = abc.ABCMeta

    @abc.abstractclassmethod
    def __init__(self, cache_class, cache_size, reader):
        self.cache_class = cache_class
        self.cache_size = cache_size
        self.reader = reader
        self.num_of_trace_elements = 0
        # self.trace_file = open(self.trace_loc, 'r')

    @abc.abstractclassmethod
    def run(self):
        pass
        # for record in self.reader():
        #     self.addOneTraceElement(record)
        # return

    # def readData(self):
    #     # this method can be overriden in the subclass
    #     line_num = 0
    #     for line in self.trace_file:
    #         yield line.strip('\n').split()[4]
    #         line_num += 1
    #         if line_num%1000 == 0:
    #             print(line_num)

    @abc.abstractclassmethod
    def addOneTraceElement(self, element):
        self.num_of_trace_elements += 1
        return

        # @abc.abstractclassmethod
        # def printResult(self):
        #     return
        #
        # @abc.abstractclassmethod
        # def returnResult(self):
        #     return
        #
        # # @abc.abstractclassmethod
        # # def showGraph(self):
        # #     raise NotImplementedError("showGraph is not implemented")
        #
        # @abc.abstractclassmethod
        # def outputResult(self, file_loc=""):
        #     return


if __name__ == "__main__":
    pA = profilerAbstract("aaa", 10, "../data/LRU.MRC")
    for record in pA.readData():
        print(record)



"""
this module provides the heatmap ploting engine, it supports both virtual time (fixed number of trace requests) and
real time, under both modes, it support using multiprocessing to do the plotting
currently, it only support LRU
the mechanism it works is it gets reuse distance from LRU profiler (parda for now)
currently supported plot type:

        # 1. hit_rate_start_time_end_time
        # 2. hit_rate_start_time_cache_size
        # 3. avg_rd_start_time_end_time
        # 4. cold_miss_count_start_time_end_time
        # 5. rd_distribution
"""

import os
import pickle
from collections import deque
from multiprocessing import Array, Process, Queue

from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.profiler.LRUProfiler import LRUProfiler
from mimircache.profiler.cHeatmap import cHeatmap
import mimircache.c_heatmap as c_heatmap
from mimircache.profiler.heatmap_subprocess import *
from mimircache.utils.printing import *
from mimircache.const import *

import matplotlib.ticker as ticker
import numpy as np
from matplotlib import pyplot as plt

# DEBUG = True

__all__ = ("run", "draw")


class heatmap:
    def __init__(self):
        # if not os.path.exists('temp/'):
        #     os.mkdir('temp')
        pass

    def _prepare_reuse_distance_and_break_points(self, mode, reader, time_interval, calculate=True, save=False,
                                                 **kwargs):
        reader.reset()

        break_points = []

        if calculate:
            # build profiler for extracting reuse distance (For LRU)
            p = LRUProfiler(reader)
            reuse_dist = p.get_reuse_distance()

            if save:
                # needs to save the data
                if not os.path.exists('temp/'):
                    os.makedirs('temp/')
                with open('temp/reuse.dat', 'wb') as ofile:
                    # pickle.dump(reuse_dist_python_list, ofile)
                    pickle.dump(reuse_dist, ofile)

        else:
            # the reuse distance has already been calculated, just load it
            with open('temp/reuse.dat', 'rb') as ifile:
                # reuse_dist_python_list = pickle.load(ifile)
                reuse_dist = pickle.load(ifile)

            breakpoints_filename = 'temp/break_points_' + mode + str(time_interval) + '.dat'
            # check whether the break points distribution table has been calculated
            if os.path.exists(breakpoints_filename):
                with open(breakpoints_filename, 'rb') as ifile:
                    break_points = pickle.load(ifile)

        # check break points are loaded or not, if not need to calculate it
        if not break_points:
            break_points = cHeatmap().gen_breakpoints(reader, mode, time_interval)
            if save:
                with open('temp/break_points_' + mode + str(time_interval) + '.dat', 'wb') as ifile:
                    pickle.dump(break_points, ifile)

        return reuse_dist, break_points

    def _prepare_multiprocess_params_LRU(self, mode, plot_type, break_points, **kwargs):
        # create the array for storing results
        # result is a dict: (x, y) -> heat, x, y is the left, lower point of heat square

        kwargs_subprocess = {}  # kw dictionary for passing parameters to subprocess
        kwargs_plot = {}  # kw dictionary passed to plotting functions
        if plot_type == "hit_rate_start_time_cache_size":
            max_rd = max(kwargs['reuse_dist'])
            kwargs_subprocess['max_rd'] = max_rd
            if 'bin_size' in kwargs:
                kwargs_subprocess["bin_size"] = kwargs['bin_size']
            else:
                kwargs_subprocess['bin_size'] = int(max_rd / DEFAULT_BIN_NUM_PROFILER)

            result = np.empty((len(break_points) - 1, max_rd // kwargs_subprocess['bin_size'] + 1), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_hit_rate_start_time_cache_size_subprocess
            kwargs_plot['yticks'] = ticker.FuncFormatter(
                lambda x, pos: '{:2.0f}'.format(kwargs_subprocess['bin_size'] * x))
            kwargs_plot['xlabel'] = 'time({})'.format(mode)
            kwargs_plot['ylabel'] = 'cache size'
            kwargs_plot['zlabel'] = 'hit rate'
            kwargs_plot['title'] = "hit rate start time cache size"

            # plt.gca().yaxis.set_major_formatter(
            #     ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(kwargs_subprocess['bin_size'] * x)))

        elif plot_type == "hit_rate_start_time_end_time":
            # (x,y) means from time x to time y
            assert 'reader' in kwargs, 'please provide reader as keyword argument'
            assert 'cache_size' in kwargs, 'please provide cache_size as keyword argument'

            reader = kwargs['reader']
            kwargs_subprocess['cache_size'] = kwargs['cache_size']
            array_len = len(break_points) - 1
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_hit_rate_start_time_end_time_subprocess

            kwargs_plot['xlabel'] = 'time({})'.format(mode)
            kwargs_plot['ylabel'] = 'time({})'.format(mode)
            kwargs_plot['yticks'] = ticker.FuncFormatter(
                lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))
            kwargs_plot['title'] = "hit_rate_start_time_end_time"

            last_access = c_heatmap.get_last_access_dist(reader.cReader)
            last_access_array = Array('l', len(last_access), lock=False)
            for i, j in enumerate(last_access):
                last_access_array[i] = j
            kwargs_subprocess['last_access_array'] = last_access_array


        elif plot_type == "avg_rd_start_time_end_time":
            # (x,y) means from time x to time y
            kwargs_plot['xlabel'] = 'time({})'.format(mode)
            kwargs_plot['ylabel'] = 'average reuse distance'
            kwargs_plot['title'] = "avg_rd_start_time_end_time"

            array_len = len(break_points) - 1
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_avg_rd_start_time_end_time_subprocess

        elif plot_type == "cold_miss_count_start_time_end_time":
            # (x,y) means from time x to time y
            kwargs_plot['xlabel'] = 'time({})'.format(mode)
            kwargs_plot['ylabel'] = 'cold miss count'
            kwargs_plot['title'] = "cold_miss_count_start_time_end_time"
            array_len = len(break_points) - 1
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_cold_miss_count_start_time_end_time_subprocess

        else:
            raise RuntimeError("cannot recognize plot type")

        kwargs_plot['xticks'] = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))

        return result, kwargs_subprocess, kwargs_plot, func_pointer

    def calculate_heatmap_dat(self, reader, mode, time_interval, plot_type,
                              algorithm="LRU", cache_params=None, num_of_threads=4, **kwargs):
        """
            calculate the data for plotting heatmap

        :param mode: mode can be either virtual time(v) or real time(r)
        :param reader: for reading the trace file
        :param plot_type: the type of heatmap to generate, possible choices are:
        1. hit_rate_start_time_end_time
        2. hit_rate_start_time_cache_size
        3. avg_rd_start_time_end_time
        4. cold_miss_count_start_time_end_time
        5. rd_distribution
        6.

        :param kwargs: possible key-arg includes the type of cache replacement algorithms(cache),

        :return: a two-dimension list, the first dimension is x, the second dimension is y, the value is the heat value
        """

        if algorithm == "LRU":
            reuse_dist, break_points = self._prepare_reuse_distance_and_break_points(
                mode, reader, time_interval, **kwargs)

            # create shared memory for child process
            reuse_dist_share_array = Array('l', len(reuse_dist), lock=False)
            for i, j in enumerate(reuse_dist):
                reuse_dist_share_array[i] = j

            result, kwargs_subprocess, kwargs_plot, func_pointer = self._prepare_multiprocess_params_LRU(mode,
                                                                                                         plot_type,
                                                                                                         break_points,
                                                                                                         reuse_dist=reuse_dist_share_array,
                                                                                                         reader=reader,
                                                                                                         **kwargs)

        else:
            assert 'cache_size' in kwargs, "you didn't provide cache_size parameter"
            if isinstance(algorithm, str):
                algorithm = cache_name_to_class(algorithm)

            # prepare break points
            if mode[0] == 'r' or mode[0] == 'v':
                break_points = cHeatmap().gen_breakpoints(reader, mode[0], time_interval)
            else:
                raise RuntimeError("unrecognized mode, it can only be r or v")

            result, kwargs_subprocess, kwargs_plot, func_pointer = self._prepare_multiprocess_params_LRU(mode,
                                                                                                         plot_type,
                                                                                                         break_points,
                                                                                                         reader=reader,
                                                                                                         **kwargs)
            kwargs_subprocess['cache_params'] = cache_params

        # create shared memory for break points
        break_points_share_array = Array('i', len(break_points), lock=False)
        for i, j in enumerate(break_points):
            break_points_share_array[i] = j

        # Jason: memory usage can be optimized by not copying whole reuse distance array in each sub process
        # Jason: change to process pool for efficiency
        map_list = deque()
        for i in range(len(break_points) - 1):
            map_list.append(i)

        q = Queue()
        process_pool = []
        process_count = 0
        result_count = 0
        map_list_pos = 0
        while result_count < len(map_list):
            if process_count < num_of_threads and map_list_pos < len(map_list):
                # LRU related heatmaps
                if algorithm == "LRU":
                    p = Process(target=func_pointer,
                                args=(map_list[map_list_pos], break_points_share_array, reuse_dist_share_array, q),
                                kwargs=kwargs_subprocess)
                else:

                    p = Process(target=calc_hit_rate_start_time_end_time_subprocess_general,
                                args=(map_list[map_list_pos], algorithm, break_points_share_array, reader, q),
                                kwargs=kwargs_subprocess)

                p.start()
                process_pool.append(p)
                process_count += 1
                map_list_pos += 1
            else:
                rl = q.get()
                for r in rl:
                    # print("[{}][{}] = {}".format(r[0], r[1], r[2]))
                    result[r[0]][r[1]] = r[2]
                process_count -= 1
                result_count += 1
            print("%2.2f%%" % (result_count / len(map_list) * 100), end='\r')
        for p in process_pool:
            p.join()

        # old 0510
        # with Pool(processes=self.num_of_threads) as p:
        #     for ret_list in p.imap_unordered(_calc_hit_rate_subprocess, map_list,
        #                                      chunksize=10):
        #         count += 1
        #         print("%2.2f%%" % (count / len(map_list) * 100), end='\r')
        #         for r in ret_list:  # l is a list of (x, y, hr)
        #             result[r[0]][r[1]] = r[2]

        return result.T, kwargs_plot


    @staticmethod
    def draw_heatmap(xydict, **kwargs):
        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = 'heatmap.png'

        masked_array = np.ma.array(xydict, mask=np.isnan(xydict))

        # print(masked_array)
        cmap = plt.cm.jet
        cmap.set_bad('w', 1.)

        if 'fixed_range' in kwargs and kwargs['fixed_range']:
            img = plt.imshow(masked_array, vmin=0, vmax=1, interpolation='nearest', origin='lower',
                             aspect='auto')
        else:
            img = plt.imshow(masked_array, interpolation='nearest', origin='lower', aspect='auto')

        cb = plt.colorbar(img)
        if 'text' in kwargs:
            (length1, length2) = masked_array.shape
            ax = plt.gca()
            ax.text(length2 // 3, length1 // 8, kwargs['text'], fontsize=20)  # , color='blue')

        if 'xlabel' in kwargs:
            plt.xlabel(kwargs['xlabel'])
        if 'ylabel' in kwargs:
            plt.ylabel(kwargs['ylabel'])
        if 'xticks' in kwargs:
            plt.gca().xaxis.set_major_formatter(kwargs['xticks'])
        if 'yticks' in kwargs:
            plt.gca().yaxis.set_major_formatter(kwargs['yticks'])
        if 'title' in kwargs:
            plt.title(kwargs['title'])

        # # # change tick from arbitrary number to real time
        # if 'change_label' in kwargs and kwargs['change_label'] and 'interval' in kwargs:
        #     ticks = ticker.FuncFormatter(lambda x, pos: '{:2.2f}'.format(x * kwargs['interval'] / (10 ** 6) / 3600))
        #     plt.gca().xaxis.set_major_formatter(ticks)
        #     plt.gca().yaxis.set_major_formatter(ticks)
        #     plt.xlabel("time/hour")
        #     plt.ylabel("time/hour")

        # plt.show()
        plt.savefig(filename, dpi=600)
        colorfulPrint("red", "plot is saved at the same directory")
        plt.clf()


    @staticmethod
    def draw_3D(xydict, **kwargs):
        from mpl_toolkits.mplot3d import Axes3D

        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = 'heatmap_3D.png'

        fig = plt.figure()
        ax = Axes3D(fig)
        # X = np.arange(-4, 4, 0.25)
        X = np.arange(0, len(xydict))
        # Y = np.arange(-4, 4, 0.25)
        Y = np.arange(0, len(xydict[0]))
        X, Y = np.meshgrid(X, Y)
        # R = np.sqrt(X ** 2 + Y ** 2)
        # Z = np.sin(R)
        Z = np.array(xydict).T

        print(X.shape)
        print(Y.shape)
        print(Z.shape)

        # print(X)
        # print(Y)
        # print(Z)

        if 'xlabel' in kwargs:
            ax.set_xlabel(kwargs['xlabel'])
        if 'ylabel' in kwargs:
            ax.set_ylabel(kwargs['ylabel'])
        if 'xticks' in kwargs:
            print("setx")
            ax.xaxis.set_major_formatter(kwargs['xticks'])
        if 'yticks' in kwargs:
            print("sety")
            ax.yaxis.set_major_formatter(kwargs['yticks'])

        ax.plot_surface(X[:, :-3], Y[:, :-3], Z[:, :-3], rstride=10000, cstride=2, cmap=plt.cm.jet)  # plt.cm.hot
        ax.contourf(X, Y, Z, zdir='z', offset=0, cmap=plt.cm.jet)
        ax.set_zlim(0, 1)
        # ax.set_zlim(-2, 2)

        plt.show()
        plt.savefig(filename, dpi=600)

    def __del_manual__(self):
        """
        cleaning
        :return:
        """
        if os.path.exists('temp/'):
            for filename in os.listdir('temp/'):
                os.remove('temp/' + filename)
            os.rmdir('temp/')

    def heatmap(self, reader, mode, time_interval, plot_type, algorithm="LRU", cache_params=None, **kwargs):
        """

        :param plot_type:
        :param mode:
        :param interval:
        :param reader:
        :param kwargs: include num_of_threads, figname
        :return:
        """
        reader.reset()

        if mode == 'r' or mode == 'v':
            xydict, kwargs_plot = self.calculate_heatmap_dat(reader, mode, time_interval, plot_type, algorithm,
                                                             cache_params=None, **kwargs)
        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)

        kwargs_plot.update(kwargs)
        self.draw_heatmap(xydict, **kwargs_plot)
        reader.reset()

        # self.__del_manual__()


def server_plot_all():
    mem_sizes = []
    with open('memSize', 'r') as ifile:
        for line in ifile:
            mem_sizes.append(int(line.strip()))

    for filename in os.listdir("../data/cloudphysics"):
        print(filename)
        if filename.endswith('.vscsitrace'):
            hm = heatmap()
            mem_size = mem_sizes[int(filename.split('_')[0][1:]) - 1] * 16
            reader = vscsiCacheReader("../data/cloudphysics/" + filename)
            if os.path.exists("0601/" + filename + "LRU_Optimal_" + str(mem_size) + "_r.png"):
                continue

            # hm.run('r', 1000000000, "hit_rate_start_time_cache_size", reader, num_of_threads=48, cache_size=mem_size,
            #        save=True, bin_size=1000, text="size = " + str(mem_size), fixed_range=True,
            #        figname='0601/' + filename + '_hit_rate_start_time_cache_size_r.png')
            # # ,change_label='False'     ,   10000000
            #
            #
            # hm.run('r', 1000000000, "avg_rd_start_time_end_time", reader, num_of_threads=48,
            #        # LRU=False, cache='optimal',
            #        calculated=True, cache_size=mem_size,
            #        figname='0601/' + filename + "_avg_rd_start_time_end_time_r.png")
            #
            #
            # hm.run('r', 1000000000, "cold_miss_count_start_time_end_time", reader, num_of_threads=48,
            #        calculated=True, cache_size=mem_size,
            #        figname='0601/' + filename + "_cold_miss_count_start_time_end_time_r.png")
            #


            del hm
            del reader



def localtest():
    CACHE_SIZE = 2000
    TIME_INTERVAL = 10000000
    reader = vscsiCacheReader("../data/trace.vscsi")
    hm = heatmap()
    from mimircache.cache.FIFO import FIFO

    # LRU
    # hm.heatmap(reader, 'r', TIME_INTERVAL, "hit_rate_start_time_end_time", num_of_threads=1, cache_size=CACHE_SIZE,
    #            save=True, text="size = " + str(CACHE_SIZE), fixed_range=True,
    #            figname='test_hit_rate_start_time_end_time_r.png')
    #
    # hm.heatmap(reader, 'r', TIME_INTERVAL, "hit_rate_start_time_cache_size", num_of_threads=8, cache_size=CACHE_SIZE,
    #            save=True, bin_size=1000, fixed_range=True,
    #            figname='test_hit_rate_start_time_cache_size_r.png')
    #
    # hm.heatmap(reader, 'r', TIME_INTERVAL, "avg_rd_start_time_end_time", num_of_threads=8,
    #            calculated=True, cache_size=CACHE_SIZE,
    #            figname="test_avg_rd_start_time_end_time_r.png")
    #
    # hm.heatmap(reader, 'r', TIME_INTERVAL, "cold_miss_count_start_time_end_time", num_of_threads=8,
    #            calculated=True, cache_size=CACHE_SIZE,
    #            figname="test_cold_miss_count_start_time_end_time_r.png")

    # Non-LRU
    hm.heatmap(reader, 'r', TIME_INTERVAL, "hit_rate_start_time_end_time", num_of_threads=8, cache_size=CACHE_SIZE,
               save=True, text="size = " + str(CACHE_SIZE), fixed_range=True, algorithm=FIFO,  # "FIFO",
               figname='test_hit_rate_start_time_end_time_r.png')



if __name__ == "__main__":
    import time

    t1 = time.time()
    localtest()
    # server_plot_all()
    # cold_miss_2d('v', 100)
    t2 = time.time()
    print(t2 - t1)




import sys
from queue import PriorityQueue

from mimircache.cache.abstractCache import cache
import mimircache.c_LRUProfiler as c_LRUProfiler
import mimircache.c_heatmap as c_heatmap

import time

class optimal(cache):
    def __init__(self, cache_size, reader):
        super().__init__(cache_size)
        reader.reset()
        self.reader = reader
        self.reader.lock.acquire()
        self.next_access = c_heatmap.get_next_access_dist(self.reader.cReader)
        self.reader.lock.release()
        # print(self.next_access)
        self.pq = PriorityQueue()
        self.seenset = set()

        # used to store the real timestamp because it is hard to del the element from priority queue
        # self.real_ts = {}
        self.ts = 0

    def get_reversed_reuse_dist(self):
        return c_LRUProfiler.get_reversed_reuse_dist(self.reader.cReader)

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.seenset:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """

        # TERRIBLE ALGORITHM HERE, NEEDS to BE CHANGED

        pq_new = PriorityQueue()
        while self.pq.qsize():
            element_t = self.pq.get()
            if element_t[1] == element:
                if self.next_access[self.ts] != -1:
                    pq_new.put((-(self.ts + self.next_access[self.ts]), element))
                else:
                    self.seenset.remove(element)
            else:
                pq_new.put(element_t)
        self.pq = pq_new



        # # update priorityQueue is hard and time-consuming (O(N)), so update the real value hash table
        # if self.next_access[element[0]] != -1:
        #     # pass
        #     self.real_ts[element[1]] = -(self.next_access[element[0]] + element[0])
        #     self.pq.put( (-(self.next_access[element[0]] + element[0]), element[1]) )
        # else:
        #     pass
        #     # self.real_ts[element[1]] = -sys.maxsize

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        # self.seenset.add(element[1])
        # self.pq.put(-(self.reversed_rd[element[0]]+element[0], element[1]))

        if self.next_access[self.ts] == -1:
            # self.pq.put((-sys.maxsize, element))
            pass
        else:
            self.pq.put((-self.next_access[self.ts] - self.ts, element))
            # print("Add: {}:{}".format(-self.next_access[element[0]]-element[0], element[1]))
            self.seenset.add(element)

    def _printCacheLine(self):
        print("size %d" % len(self.seenset))
        for i in self.seenset:
            print(i, end='\t')
        print('')

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        # needs to change this for performance
        element = self.pq.get()
        self.seenset.remove(element[1])
        # print("evicting "+str(element))
        return element




        # element = self.pq.get()
        # print('evict: {}?: '.format(element[1]), end='\t')
        # self.seenset.remove(element[1])
        # if not element[1] in self.real_ts:
        #     print(element[1])
        # while element[1] in self.real_ts:
        #     if self.real_ts[element[1]] != element[0]:
        #         print("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA")
        #         element = self.pq.get()
        #     else:
        #         print('BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB')
        #         # del self.real_ts[element[1]]
        #         print(element[1])
        #         break

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not,
                        !!! Attention, for optimal, the element is a tuple of
                        (timestamp, real_request)
        :return: True if element in the cache
        """
        if self.pq.qsize() != len(self.seenset):
            print("ERROR: %d: %d" % (self.pq.qsize(), len(self.seenset)))
            print(self.seenset)
            sys.exit(1)
        if self.checkElement(element):
            self._updateElement(element)
            self.ts += 1
            return True
        else:
            self._insertElement(element)
            if self.pq.qsize() > self.cache_size:
                self._evictOneElement()
            self.ts += 1
            return False

    def __repr__(self):
        return "Optimal Cache, current size: {}".format(self.cache_size, len(self.seenset),
                                                        super().__repr__())


if __name__ == "__main__":
    from mimircache.cacheReader.plainReader import plainCacheReader
    from mimircache.cacheReader.vscsiReader import vscsiCacheReader
    import mimircache.c_cacheReader as c_cacheReader

    reader = plainCacheReader('../data/test4.dat')
    # reader = vscsiCacheReader('../data/trace.vscsi')
    # for i in reader:
    #     print(i)

    # reader.reset()
    print(reader.get_num_of_total_requests())

    # for i in c_LRUProfiler.get_reuse_dist_seq(reader.cReader):
    #     print(i)
    #     pass

    num = 0
    c = optimal(100, reader)
    hit = 0
    miss = 0

    for i in reader:
        if c.addElement(i):
            hit += 1
        else:
            miss += 1
        # c.printCacheLine()
        # print(c.real_ts)
        # print("")
        num += 1
    print("hit: %d, hit rate: %f" % (hit, hit / (hit + miss)))



import math

from mimircache.utils.LinkedList import LinkedList


class modifiedLRU:
    def __init__(self, size=1000):
        self.max_size = size
        # this size includes holes, if want real element size use self.cacheLinkedList.size
        self.size = 0
        # self.cache   = OrderedDict()
        self.cacheLinkedList = LinkedList()
        self.cacheDict = dict()
        self.headPos = 1
        self.tailPos = 1
        self.numOfHoles = 0
        self.holePosSet = set()

    def checkElement(self, content):
        """
        :param content: the content for search
        :return: whether the given element is in the cache
        """
        if content in self.cacheDict:
            return True
        else:
            return False

    def getRank(self, content):
        """

        :param content: the content of element, which is also the key in cache_dict
        :return: rank if in the cache, otherwise -1
        """
        return self.cacheDict.get(content, (-1, -1))[0]

    def __updateElement__(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: original rank
        """
        rank = self.cacheDict.get(element, (-1, -1))[0] - self.headPos  # this may have holes
        rank -= math.floor(rank / self.size * self.numOfHoles)  # adjust
        if rank < 0:
            print("waht, rank<0???\t" + str(self.cacheDict.get(element, (-1, -1))[0]) + '\t' + str(self.headPos) + '\t')
            print(str(rank) + "\t" + str(self.cacheDict.get(element, (-1, -1))[0] - self.headPos) + '\t'
                  + str(self.size) + '\t' + str(self.numOfHoles))

        self.headPos -= 1
        self.cacheDict[element][0] = self.headPos
        self.numOfHoles += 1
        self.size += 1
        self.holePosSet.add(rank)

        self.cacheLinkedList.moveNodeToHead(self.cacheDict[element][1])
        return rank

    def __insertElement__(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        node = self.cacheLinkedList.insertAtHead(element)
        self.headPos -= 1
        self.cacheDict[element] = [self.headPos, node]
        self.size += 1

        if self.cacheLinkedList.size > self.max_size:
            self.__evictOneElement__()
            self.tailPos -= 1

    def printLinkedList(self):
        for i in self.cacheLinkedList:
            print(i.content)

    def __evictOneElement__(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        content = self.cacheLinkedList.removeFromTail()
        del self.cacheDict[content]

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: -1 if not in cache, otherwise old rank
        """
        if self.checkElement(element):
            return self.__updateElement__(element)
        else:
            self.__insertElement__(element)
            return -1

    def reArrange(self):
        pass



import math

from mimircache.profiler.abstract.abstractLRUProfiler import abstractLRUProfiler


class basicLRUProfiler(abstractLRUProfiler):
    def __init__(self, cache_class, cache_size, bin_size, reader):
        super().__init__(cache_class, cache_size, bin_size, reader)

    def run(self):
        for record in self.reader:
            self.addOneTraceElement(record)
        self.calculate()

    def addOneTraceElement(self, element):
        super().addOneTraceElement(element)
        rank = self.cache.addElement(element)
        if rank != -1:
            self.HRC_bin[math.ceil(rank / self.bin_size)] += 1


if __name__ == "__main__":
    pass



import math

from mimircache.cache.ARC import ARC
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.Random import Random
from mimircache.cache.SLRU import SLRU
from mimircache.cache.FIFO import FIFO
from mimircache.cache.Optimal import optimal

import mimircache.c_generalProfiler as c_generalProfiler

def calc_hit_rate_start_time_end_time_subprocess_general(order, cache, break_points_share_array, reader, q, **kwargs):
    """
    the child process for calculating hit rate for a general cache replacement algorithm,
    each child process will calculate for a column with fixed starting time
    :param cache:
    :param order: the order of column the child is working on
    :param break_points_share_array:
    :param reader:
    :param q
    :return: nothing, but add to the queue a list of result in the form of (x, y, hit_rate) with x as fixed value
    """

    # new for optimal using c_generalProfiler
    # c_generalProfiler.get_hit_rate()

    cache_size = kwargs['cache_size']
    # if cache == 'Random':
    #     c = Random(cache_size=cache_size)
    # elif cache == 'SLRU':
    #     c = SLRU(cache_size=cache_size)
    # elif cache == 'AdaptiveSLRU':
    #     c = AdaptiveSLRU(cache_size=cache_size)
    # elif cache == 'LFU_RR':
    #     c = LFU_RR(cache_size=cache_size)
    # elif cache == 'LRU':
    #     c = LRU(cache_size=cache_size)
    # elif cache == "optimal":
    #     c = optimal(cache_size, reader)
    # elif cache == 'FIFO':
    #     c = FIFO(cache_size=cache_size)
    # else:
    if 'cache_params' in kwargs and kwargs['cache_params']:
        c = cache(cache_size=cache_size, cache_params=kwargs['cache_params'])
    else:
        c = cache(cache_size=cache_size)


    result_list = []
    total_hc = 0  # total hit count
    total_mc = 0  # total miss count
    pos_in_break_points = order + 1
    line_num = 0
    # if type(reader) != plainCacheReader:
    #     reader_new = plainCacheReader('temp.dat')
    # else:
    #     reader_new = type(reader)(reader.file_loc)

    reader_new = type(reader)(reader.file_loc)

    # TODO: figure out line size here and add seek method in reader base class
    # TODO: use mmap here to improve performance
    for line in reader_new:
        if line_num < break_points_share_array[order]:
            line_num += 1
            continue
        # fix this hack
        if cache == 'optimal':
            c.ts = line_num

        line_num += 1
        if c.addElement(line):
            total_hc += 1
        else:
            total_mc += 1

        if pos_in_break_points < len(break_points_share_array) and \
                        line_num == break_points_share_array[pos_in_break_points]:
            hr = total_hc / (break_points_share_array[pos_in_break_points] - break_points_share_array[order])
            result_list.append((order, pos_in_break_points - 1, hr))
            pos_in_break_points += 1
            # print("{}: {}".format(total_hc, total_mc))
    q.put(result_list)
    reader_new.close()
    # return result_list


# LRU

def calc_hit_rate_start_time_cache_size_subprocess(order, break_points_share_array, reuse_dist_share_array, q,
                                                   **kwargs):
    """
    the child process for calculating hit rate of different cache size with different starting time, but fixed end time,
    and each child process will calculate for a column with a given starting time
    :param q:
    :param reuse_dist_share_array:
    :param break_points_share_array:
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, hit_rate) with x as fixed value(starting time), y as cache size
    """

    max_rd = kwargs['max_rd']
    bin_size = kwargs['bin_size']

    result_list = []

    rd_distribution = [0] * int(max_rd // bin_size + 1)

    for i in range(break_points_share_array[order], break_points_share_array[-1]):
        rd_distribution[reuse_dist_share_array[i] // bin_size] += 1

    num_of_total_request = break_points_share_array[-1] - break_points_share_array[order]

    accum = 0
    for i in range(len(rd_distribution)):
        accum += rd_distribution[i]
        result_list.append((order, i, accum / num_of_total_request))

    q.put(result_list)


def _hit_rate_start_time_end_time_calc_hit_count(reuse_dist_array, last_access_array, cache_size, begin_pos, end_pos,
                                                 real_start, **kwargs):
    """
    called by hit_rate_start_time_end_time to calculate hit count
    :rtype: count of hit
    :param reuse_dist_array:
    :param cache_size:
    :param begin_pos:
    :param end_pos: end pos of trace in current partition (not included)
    :param real_start: the real start position of cache trace
    :return:
    """

    hit_count = 0
    miss_count = 0
    try:
        for i in range(begin_pos, end_pos):
            if reuse_dist_array[i] == -1:
                # never appear
                miss_count += 1
                continue
            if last_access_array[i] - (i - real_start) <= 0 and reuse_dist_array[i] < cache_size:
                # hit
                hit_count += 1
            else:
                # miss
                miss_count += 1
    except:
        print("error: begin: {}, end: {}, i: {}".format(begin_pos, end_pos, i))
    return hit_count


def calc_hit_rate_start_time_end_time_subprocess(order, break_points_share_array, reuse_dist_share_array, q,
                                                 **kwargs):
    """
    the child process for calculating hit rate, each child process will calculate for
    a column with fixed starting time
    :param q:
    :param reuse_dist_share_array:
    :param break_points_share_array:
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, hit_rate) with x as fixed value
    """

    cache_size = kwargs['cache_size']
    last_access_array = kwargs['last_access_array']

    result_list = []
    total_hc = 0
    for i in range(order + 1, len(break_points_share_array)):
        hc = _hit_rate_start_time_end_time_calc_hit_count(reuse_dist_share_array, last_access_array, cache_size,
                                                          break_points_share_array[i - 1],
                                                          break_points_share_array[i],
                                                          break_points_share_array[order])
        total_hc += hc
        hr = total_hc / (break_points_share_array[i] - break_points_share_array[order])
        result_list.append((order, i - 1, hr))
    q.put(result_list)


def calc_avg_rd_start_time_end_time_subprocess(order, break_points_share_array, reuse_dist_share_array, q, **kwargs):
    """
    the child process for calculating average reuse distance in each block, each child process will calculate for
    a column with fixed starting time
    :param q:
    :param reuse_dist_share_array:
    :param break_points_share_array:
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, hit_rate) with x as fixed value
    """

    result_list = []
    rd = 0
    never_see = 0
    count = 0
    for i in range(order + 1, len(break_points_share_array)):
        for j in range(break_points_share_array[i - 1], break_points_share_array[i]):
            if reuse_dist_share_array[j] != -1:
                rd += reuse_dist_share_array[j]
                count += 1
            else:
                never_see += 1
        if break_points_share_array[i] - break_points_share_array[order] - never_see != 0:
            result_list.append(
                (order, i - 1, rd / (break_points_share_array[i] - break_points_share_array[order] - never_see)))
    q.put(result_list)


def calc_cold_miss_count_start_time_end_time_subprocess(order, break_points_share_array, reuse_dist_share_array, q,
                                                        **kwargs):
    """
    the child process for calculating cold miss count in each block, each child process will calculate for
    a column with fixed starting time
    :param q:
    :param reuse_dist_share_array:
    :param break_points_share_array:
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, miss_count) with x as fixed value
    """

    result_list = []
    rd = 0
    never_see = 0
    count = 0
    for i in range(order + 1, len(break_points_share_array)):
        for j in range(break_points_share_array[i - 1], break_points_share_array[i]):
            if reuse_dist_share_array[j] != -1:
                rd += reuse_dist_share_array[j]
                count += 1
            else:
                never_see += 1

        result_list.append((order, i - 1, never_see))

    q.put(result_list)


def calc_rd_distribution_subprocess_deprecated(order, break_points_share_array, reuse_dist_share_array, q, **kwargs):
    """
    deprecated
    For LRU
    the child process for calculating average reuse distance in each block, each child process will calculate for
    a column with fixed starting time
    :param q:
    :param reuse_dist_share_array:
    :param break_points_share_array:
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, hit_rate) with x as fixed value
    """

    LOG_NUM = kwargs['log_num']
    max_rd = kwargs['max_rd']

    result_list = []
    rd_bucket = [0] * int(math.log(max_rd, LOG_NUM) + 1)

    for j in range(break_points_share_array[order], break_points_share_array[order + 1]):
        if reuse_dist_share_array[j] != -1:
            if reuse_dist_share_array[j] == 0:
                rd_bucket[0] += 1
            else:
                rd_bucket[int(math.log(reuse_dist_share_array[j], LOG_NUM))] += 1

    for i in range(len(rd_bucket)):
        result_list.append((order, i - 1, rd_bucket[i]))

    q.put(result_list)



def calc_cold_miss_count(break_points, reader, **kwargs):
    """
    the function for calculating cold miss count in each block
    :param reader:
    :param break_points:
    :return:
    """

    result_list = [0] * (len(break_points) - 1)
    seen_set = set()
    for i in range(len(break_points) - 1):
        never_see = 0
        for j in range(break_points[i + 1] - break_points[i]):
            r = next(reader)
            if r not in seen_set:
                seen_set.add(r)
                never_see += 1
        result_list[i] = never_see

    return result_list



from collections import OrderedDict

from mimircache.cache.abstractLFU import abstractLFU


class LFU_MRU(abstractLFU):
    def __init__(self, cache_size=1000):
        super(LFU_MRU, self).__init__(cache_size)
        self.cacheDict = OrderedDict()

    def find_evict_key(self):
        evict_key = self.least_freq_elements_list.pop()
        return evict_key

    def _updateElement(self, element):
        super()._updateElement(element)
        self.cacheDict.move_to_end(element)



from collections import defaultdict

from mimircache.cache.abstractCache import cache
from mimircache.utils.LinkedList import LinkedList


class ARC(cache):
    def __init__(self, cache_size=1000, p=0.5, ghostlist_size=-1):
        """

        :param cache_size:
        :param p: the position to separate cache for LRU1 and LRU2
        :param ghostlist_size:
        :return:
        """
        super().__init__(cache_size)
        self.p = p
        if ghostlist_size == -1:
            self.ghostlist_size = self.cache_size

        # dict/list2 is for referenced more than once
        self.linkedList1 = LinkedList()
        self.linkedList2 = LinkedList()
        self.lru_list_head_p1 = None
        self.lru_list_head_p2 = None

        self.cacheDict1 = dict()  # key -> linked list node (in reality, it should also contains value)
        self.cacheDict2 = dict()  # key -> linked list node (in reality, it should also contains value)
        self.cacheDict1_ghost = defaultdict(
            lambda: 0)  # key -> linked list node (in reality, it should also contains value)
        self.cacheDict2_ghost = defaultdict(
            lambda: 0)  # key -> linked list node (in reality, it should also contains value)

    def checkElement(self, element):
        """
        :param element: the element for search
        :return: whether the given element is in the cache
        """
        if element in self.cacheDict1 or element in self.cacheDict2:
            return True
        else:
            return False

    def check_ghost_list(self, element):
        """
        :param element: the element for search
        :return: whether the given element is in the cache
        """
        if element in self.cacheDict1_ghost or element in self.cacheDict2_ghost:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """
        if element in self.cacheDict1:
            # move to part2
            # get the node and remove from part1, size of part1 reduced by 1
            node = self.cacheDict1[element]
            # print("before remove")
            # for i in self.linkedList1:
            #     print(i.content, end='\t')
            # print("")
            if node == self.lru_list_head_p1:
                self.lru_list_head_p1 = self.lru_list_head_p1.next
            self.linkedList1.removeNode(node)
            # print("after remove")
            # for i in self.linkedList1:
            #     print(i.content, end='\t')
            # print("")

            del self.cacheDict1[element]

            # insert into part2
            self.linkedList2.insertNodeAtTail(node)
            self.cacheDict2[node.content] = node
            # delete one from part1, insert one into part2, the total size should not change, just the balance changes
            # check whether lru_list_head_p2 has been initialized or not
            if not self.lru_list_head_p2:
                self.lru_list_head_p2 = node


        else:
            node = self.cacheDict2[element]
            if node == self.lru_list_head_p2 and node.next:
                self.lru_list_head_p2 = self.lru_list_head_p2.next

            self.linkedList2.moveNodeToTail(node)

    def _insertElement(self, element):
        '''
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: evicted element or None
        '''
        return_content = None
        if self.linkedList1.size + self.linkedList2.size >= self.cache_size:
            # needs to evict one element, depend on ghost list to decide evict from part1 or part2
            return_content = self._evictOneElement(element)

        # insert into part 1
        node = self.linkedList1.insertAtTail(element)
        self.cacheDict1[element] = node
        if not self.lru_list_head_p1:
            self.lru_list_head_p1 = node
        return return_content

    def _printCacheLine(self):
        print('list 1(including ghost list): ')
        for i in self.linkedList1:
            try:
                print(i.content, end='\t')
            except:
                print(i.content)
        print('\nlist 2(including ghost list): ')
        for i in self.linkedList2:
            try:
                print(i.content, end='\t')
            except:
                print(i.content)
        print(' ')

    def _evictOneElement(self, element):
        """
        evict one element from the cache line into ghost list, then check ghost list,
        if oversize, then evict one from ghost list
        :param: element: the missed request
        :return: content of element evicted into ghost list
        """
        return_content = None
        if element in self.cacheDict1_ghost and len(self.cacheDict2) > 0:
            # evict one from part2 LRU, add into ghost list
            return_content = content = self.lru_list_head_p2.content
            del self.cacheDict2[content]
            self.lru_list_head_p2 = self.lru_list_head_p2.next
            self.cacheDict2_ghost[content] += 1

            if (self.linkedList2.size - len(self.cacheDict2)) > self.ghostlist_size:
                # the first part is the size of ghost list of part2
                content = self.linkedList2.removeFromHead()
                if self.cacheDict2_ghost[content] == 1:
                    del self.cacheDict2_ghost[content]
                else:
                    self.cacheDict2_ghost[content] -= 1
                assert (self.linkedList2.size - len(self.cacheDict2)) == self.ghostlist_size


        elif element in self.cacheDict2_ghost and len(self.cacheDict1) > 0:
            # evict one from part1 LRU, add into ghost list
            return_content = content = self.lru_list_head_p1.content
            del self.cacheDict1[content]

            self.lru_list_head_p1 = self.lru_list_head_p1.next
            self.cacheDict1_ghost[content] += 1

            if (self.linkedList1.size - len(self.cacheDict1)) > self.ghostlist_size:
                # the first part is the size of ghost list of part1
                content = self.linkedList1.removeFromHead()
                if self.cacheDict1_ghost[content] == 1:
                    del self.cacheDict1_ghost[content]
                else:
                    self.cacheDict1_ghost[content] -= 1
                assert (self.linkedList1.size - len(self.cacheDict1)) == self.ghostlist_size

        else:
            # not in any ghost list, check p value
            if len(self.cacheDict2) == 0 or len(self.cacheDict1) / len(self.cacheDict2) > self.p:
                # remove from part1
                return_content = content = self.lru_list_head_p1.content
                del self.cacheDict1[content]

                self.lru_list_head_p1 = self.lru_list_head_p1.next
                self.cacheDict1_ghost[content] += 1

                if (self.linkedList1.size - len(self.cacheDict1)) > self.ghostlist_size:
                    # the first part is the size of ghost list of part1
                    content = self.linkedList1.removeFromHead()
                    if self.cacheDict1_ghost[content] == 1:
                        del self.cacheDict1_ghost[content]
                    else:
                        self.cacheDict1_ghost[content] -= 1
                    assert (self.linkedList1.size - len(self.cacheDict1)) == self.ghostlist_size
            else:
                # remove from part2
                return_content = content = self.lru_list_head_p2.content
                del self.cacheDict2[content]

                self.lru_list_head_p2 = self.lru_list_head_p2.next
                self.cacheDict2_ghost[content] += 1

                if (self.linkedList2.size - len(self.cacheDict2)) > self.ghostlist_size:
                    # the first part is the size of ghost list of part2
                    content = self.linkedList2.removeFromHead()
                    if self.cacheDict2_ghost[content] == 1:
                        del self.cacheDict2_ghost[content]
                    else:
                        self.cacheDict2_ghost[content] -= 1
                    assert (self.linkedList2.size - len(self.cacheDict2)) == self.ghostlist_size

        return return_content

    # for debug
    def _check_lru_list_p(self):
        size1 = sum(i for i in self.cacheDict1_ghost.values())
        node1 = self.linkedList1.head.next
        for i in range(size1):
            node1 = node1.next
        assert node1 == self.lru_list_head_p1, "LRU list1 head pointer wrong"

        size2 = sum(i for i in self.cacheDict2_ghost.values())
        node2 = self.linkedList2.head.next
        for i in range(size2):
            node2 = node2.next
        print(size2)
        if node2:
            print(node2.content)
        else:
            print(node2)
        if self.lru_list_head_p2:
            print(self.lru_list_head_p2.content)
        else:
            print(self.lru_list_head_p2)
        assert node2 == self.lru_list_head_p2, "LRU list2 head pointer wrong"

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: None
        """
        if self.checkElement(element):
            self._updateElement(element)
            # self.printCacheLine()
            # self._check_lru_list_p()
            return True
        else:
            self._insertElement(element)
            # self.printCacheLine()
            # self._check_lru_list_p()
            return False

    def __repr__(self):
        return "ARC, given size: {}, current part1 size: {}, part2 size: {}".format(
            self.cache_size, self.linkedList1.size, self.linkedList2.size)



import os
import logging
import mimircache.c_LRUProfiler as c_LRUProfiler

from mimircache.const import *
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.cacheReader.abstractReader import cacheReaderAbstract
import matplotlib.pyplot as plt
from mimircache.utils.printing import *

class LRUProfiler:
    def __init__(self, reader, cache_size=-1):
        self.cache_size = cache_size
        self.reader = reader

        assert isinstance(reader, cacheReaderAbstract), "you provided an invalid cacheReader: {}".format(reader)

        # if the given file is not basic reader, needs conversion
        need_convert = True
        for instance in c_available_cacheReader:
            if isinstance(reader, instance):
                need_convert = False
                break
        if need_convert:
            self.prepare_file()
        self.num_of_lines = self.reader.get_num_of_total_requests()

    all = ["get_hit_count", "get_hit_rate", "get_miss_rate", "get_reuse_distance",
           "plotMRC", "plotHRC", "get_best_cache_sizes"]

    def prepare_file_remove_one(self):
        """
        this function will prepare the file, meanwhile remove the request that appear only once
        :return:
        """
        self.num_of_lines = 0
        logging.debug("changing file format")
        seen_dict = {}
        for e in self.reader:
            seen_dict[e] = seen_dict.get(e, 0) + 1
        self.reader.reset()
        print(len(seen_dict))
        with open('temp.dat', 'w') as ofile:
            j = self.reader.read_one_element()
            while j is not None:
                self.num_of_lines += 1
                if seen_dict[j] > 1:
                    ofile.write(str(j) + '\n')
                j = self.reader.read_one_element()
        self.reader = plainCacheReader('temp.dat')
        print(self.num_of_lines)

    def prepare_file(self):
        self.num_of_lines = 0
        with open('temp.dat', 'w') as ofile:
            i = self.reader.read_one_element()
            while i is not None:
                self.num_of_lines += 1
                ofile.write(str(i) + '\n')
                i = self.reader.read_one_element()
        self.reader = plainCacheReader('temp.dat')

    def addOneTraceElement(self, element):
        # do not need this function in this profiler
        pass

    def _kwargs_parse(self, **kwargs):
        if 'begin' in kwargs:
            begin = kwargs['begin']
        if 'end' in kwargs:
            end = kwargs['end']
        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']


    def get_hit_count(self, **kargs):
        if 'cache_size' not in kargs:
            kargs['cache_size'] = self.cache_size
        hit_count = c_LRUProfiler.get_hit_count_seq(self.reader.cReader, **kargs)
        return hit_count

    def get_hit_rate(self, **kargs):
        if 'cache_size' not in kargs:
            kargs['cache_size'] = self.cache_size
        hit_rate = c_LRUProfiler.get_hit_rate_seq(self.reader.cReader, **kargs)
        return hit_rate

    def get_miss_rate(self, **kargs):
        if 'cache_size' not in kargs:
            kargs['cache_size'] = self.cache_size
        miss_rate = c_LRUProfiler.get_miss_rate_seq(self.reader.cReader, **kargs)
        return miss_rate

    def get_reuse_distance(self, **kargs):
        rd = c_LRUProfiler.get_reuse_dist_seq(self.reader.cReader, **kargs)
        return rd

    def plotMRC(self, figname="MRC.png", threshhold=0.98, **kwargs):
        MRC = self.get_miss_rate(**kwargs)
        try:
            stop_point = len(MRC) - 2
            if self.cache_size == -1 and 'cache_size' not in kwargs:
                for i in range(len(MRC) - 3, 0, -1):
                    if MRC[i] >= MRC[-3] / threshhold:
                        stop_point = i
                        break
                if stop_point + 200 < len(MRC) - 2:
                    stop_point += 200
                else:
                    stop_point = len(MRC) - 2

            plt.plot(MRC[:stop_point])
            plt.xlabel("cache Size")
            plt.ylabel("Miss Rate")
            plt.title('Miss Rate Curve', fontsize=18, color='black')
            plt.savefig(figname, dpi=300)
            colorfulPrint("red", "plot is saved at the same directory")
            plt.show()
            plt.clf()
            return stop_point
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def plotHRC(self, figname="HRC.png", threshhold=0.98, **kwargs):
        HRC = self.get_hit_rate(**kwargs)
        try:
            stop_point = len(HRC) - 2
            if self.cache_size == -1 and 'cache_size' not in kwargs:
                for i in range(len(HRC) - 3, 0, -1):
                    if HRC[i] <= HRC[-3] * threshhold:
                        stop_point = i
                        break
                if stop_point + 200 < len(HRC) - 2:
                    stop_point += 200
                else:
                    stop_point = len(HRC) - 2
            # print(HRC)
            # print("stop: {}, HRC[-3]: {}, ".format(stop_point, HRC[-3]))
            plt.xlim(0, stop_point)
            plt.plot(HRC[:stop_point])
            plt.xlabel("cache Size")
            plt.ylabel("Hit Rate")
            plt.title('Hit Rate Curve', fontsize=18, color='black')
            plt.savefig(figname, dpi=600)
            colorfulPrint("red", "plot is saved at the same directory")

            plt.show()
            plt.clf()
            return stop_point
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def __del__(self):
        if (os.path.exists('temp.dat')):
            os.remove('temp.dat')


    def get_best_cache_sizes(self, num, force_spacing=200, cut_off_divider=20):
        best_cache_sizes = c_LRUProfiler.get_best_cache_sizes(self.reader.cReader, num, force_spacing, cut_off_divider)
        return best_cache_sizes


def _plot_HRC(filepath):
    reader = vscsiCacheReader(filepath)
    print("begin plotting " + filepath)
    cache_size = max(LRUProfiler(reader, cache_size=-1).get_best_cache_sizes(20, 200, 10))
    # LRUProfiler(reader, cache_size=cache_size).get_hit_count()
    # print('after hit count')
    LRUProfiler(reader, cache_size=int(cache_size * 1.5)).plotHRC(
        figname='0625_HRC/' + filepath.split('/')[-1] + '_HRC.png')
    reader.close()
    print("finish plotting " + filepath)


def process_test(a):
    print("get " + str(a))


def _server_plot_all(path, threads):
    import os
    from multiprocessing import Pool

    file_list = []
    for filename in os.listdir(path):
        if filename.endswith('.vscsitrace'):
            figname = 'HRC_' + filename + '.png'
            if os.path.exists('0625_HRC/' + figname):
                continue
            else:
                file_list.append(path + '/' + filename)
        else:
            print(filename)

    print(file_list)
    p = Pool(processes=threads)
    r = p.imap_unordered(_plot_HRC, file_list)
    print('after pool')
    p.close()
    p.join()

    # for i in file_list:
    #     _plot_HRC(i)


if __name__ == "__main__":
    # p = pardaProfiler(30000, plainCacheReader("../data/parda.trace"))
    # p = pardaProfiler(30000, csvCacheReader("../data/trace_CloudPhysics_txt", 4))
    # p = parda(LRU, 3000000, basicCacheReader("temp.dat"))
    # p.run(parda_mode.seq, threads=4)
    # p.get_reuse_distance()
    import os
    import time

    # from pympler.tracker import SummaryTracker

    # tracker = SummaryTracker()

    t1 = time.time()
    # for i in range(6):
    # reader = vscsiCacheReader("/run/shm/traces/w106_vscsi1.vscsitrace") # ""../data/trace.vscsi")
    reader = vscsiCacheReader("../data/trace.vscsi")
    p = LRUProfiler(reader)
    rd_a = p.get_reuse_distance()
    print(rd_a)

    hrs = p.get_hit_rate()
    with open('HR', 'w') as ofile:
        for hr in hrs:
            ofile.write("{}\n".format(hr))

    # print(p.get_best_cache_sizes(20, 600, 10))

    # _server_plot_all('/run/shm/traces/', 48)

    t2 = time.time()

    print(t2 - t1)

    # p = LRUProfiler(vscsiCacheReader(" /run/shm/traces/w106_vscsi1.vscsitrace"), 10000)
    # p.plotHRC()

    # print(len())
    # count = 0
    # for i in rd_a:
    #     if i==0:
    #         count+=1
    # print(count)

    # tracker.print_diff()


    # p = pardaProfiler(2000, csvCacheReader('../data/mining/mining.dat.original', 1))
    # p = pardaProfiler(2000, plainCacheReader('../data/mining/mining.dat.original'))
    # p._test()


    # for f in os.listdir('../data/mining/'):
    #     shutil.copy('../data/mining/' + f, '../data/mining/mining.dat')
    #     print(f)
    #     p._test()
    # p.run_with_specified_lines(10000, 20000)
    # p.plotHRC(autosize=True, autosize_threshhold=0.00001)






def _cal_log_num(self, max_rd, length):
    """
    find out the LOG_NUM that makes the number of pixels in y axis to match the number of pixels in x axis
    :param max_rd: maxinum reuse distance
    :param length: length of break points (also number of pixels along x axis)
    :return:
    """
    init_log_num = 10
    l2_prev = length
    while True:
        l2 = math.log(max_rd, init_log_num)
        if l2 > length > l2_prev:
            return init_log_num
        l2_prev = l2
        init_log_num = (init_log_num - 1) / 2 + 1


@staticmethod
def _get_breakpoints_virtualtime(bin_size, total_length):
    """

    :param bin_size: the size of the group (virtual time slice)
    :param total_length: the total length of trace requests
    :return: a list of break points

    """
    break_points = []
    assert total_length // bin_size > 10, "bin size too large, please choose a smaller one, " \
                                          "total length %d, bin size %d" % (total_length, bin_size)
    for i in range(total_length // bin_size):
        break_points.append(i * bin_size)
    if break_points[-1] != total_length - 1:
        break_points.append(total_length - 1)

    if (len(break_points)) > 10000:
        colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
        exact size: %d, it may take a very long time, if you didn't intend to do it, \
        please try with a larger time stamp" % len(break_points))

    return break_points


@staticmethod
def _get_breakpoints_realtime(reader, interval):
    """
    given reader(vscsi) and time interval, split the data according to the time interval,
    save the order into break_points(list)
    :param reader:
    :param interval:
    :return:
    """
    reader.reset()
    break_points = []
    prev = 0
    for num, line in enumerate(reader.lines()):
        if (float(line[0]) - prev) > interval:
            break_points.append(num)
            prev = float(line[0])

    # noinspection PyUnboundLocalVariable
    if line[0] != prev:
        # noinspection PyUnboundLocalVariable
        break_points.append(num)

    if (len(break_points)) > 10000:
        colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
        exact size: %d, it may take a very long time, if you didn't intend to do it, \
        please try with a larger time stamp" % len(break_points))

    # print(break_points)
    return break_points


@staticmethod
def _calc_hit_rate(reuse_dist_array, last_access_dist_array, cache_size, begin_pos, end_pos):
    """
    the original function for calculating hit rate, given the reuse_dist_array, cache_size,
    begin position and end position in the reuse_dist_array, return hit rate
    now isolate as a separate function
    :param reuse_dist_array:
    :param cache_size:
    :param begin_pos:
    :param end_pos:
    :return:
    """
    hit_count = 0
    miss_count = 0
    for i in range(begin_pos, end_pos):
        if reuse_dist_array[i] == -1:
            # never appear
            miss_count += 1
            continue
        if last_access_dist_array[i] - (i - begin_pos) <= 0 and reuse_dist_array[i] < cache_size:
            # hit
            hit_count += 1
            print("{}: {}".format(i, reuse_dist_array[i]))
        else:
            # miss
            miss_count += 1
    print("hit={}, hit+miss={}, total size:{}, hit rate:{}".format(hit_count, hit_count + miss_count,
                                                                   end_pos - begin_pos,
                                                                   hit_count / (end_pos - begin_pos)))
    return hit_count / (end_pos - begin_pos)




"""
this module provides the heatmap ploting engine, it supports both virtual time (fixed number of trace requests) and
real time, under both modes, it support using multiprocessing to do the plotting
currently, it only support LRU
the mechanism it works is it gets reuse distance from LRU profiler (parda for now)
currently supported plot type:

        # 1. hit_rate_start_time_end_time
        # 2. hit_rate_start_time_cache_size
        # 3. avg_rd_start_time_end_time
        # 4. cold_miss_count_start_time_end_time
        # 5. rd_distribution
"""

import logging
import os
import pickle
from collections import deque
from multiprocessing import Array, Process, Queue

import matplotlib
import matplotlib.ticker as ticker
import mimircache.c_heatmap as c_heatmap
import numpy as np
from matplotlib import pyplot as plt

from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.oldModule.heatmap_subprocess2d import *
from mimircache.oldModule.pardaProfiler import pardaProfiler
from mimircache.profiler.LRUProfiler import LRUProfiler
from mimircache.profiler.heatmap_subprocess import *
from mimircache.utils.printing import *

DEBUG = True

__all__ = ("run", "draw", "prepare_heatmap_dat")


# heatmap gradient, only works in _calc_rd_count_subprocess
# HEATMAP_GRADIENT = 10000
# LOG_NUM = 0




class heatmap:
    def __init__(self, cache_class=LRU):
        self.cache_class = cache_class
        if not os.path.exists('temp/'):
            os.mkdir('temp')

    def _prepare_reuse_distance_and_break_points(self, mode, reader, calculate, save=False):
        # assert isinstance(reader, vscsiCacheReader), "Currently only supports vscsiReader"
        reader.reset()

        break_points = []

        if calculate:
            # build profiler for extracting reuse distance (For LRU)
            # p = pardaProfiler(30000, reader)
            p = LRUProfiler(30000, reader)
            # reuse distance, c_reuse_dist_long_array is an array in dtype C long type
            # c_reuse_dist_long_array = p.get_reuse_distance()
            reuse_dist = p.get_reuse_distance()

            if save:
                # needs to save the data
                # for i in c_reuse_dist_long_array:
                #     reuse_dist_python_list.append(i)

                if not os.path.exists('temp/'):
                    os.makedirs('temp/')
                with open('temp/reuse.dat', 'wb') as ofile:
                    # pickle.dump(reuse_dist_python_list, ofile)
                    pickle.dump(reuse_dist, ofile)

        else:
            # the reuse distance has already been calculated, just load it
            with open('temp/reuse.dat', 'rb') as ifile:
                # reuse_dist_python_list = pickle.load(ifile)
                reuse_dist = pickle.load(ifile)

            breakpoints_filename = 'temp/break_points_' + mode + str(self.interval) + '.dat'
            # check whether the break points distribution table has been calculated
            if os.path.exists(breakpoints_filename):
                with open(breakpoints_filename, 'rb') as ifile:
                    break_points = pickle.load(ifile)

        # check break points are loaded or not, if not need to calculate it
        if not break_points:
            if mode == 'r':
                break_points = self._get_breakpoints_realtime(reader, self.interval)
            elif mode == 'v':
                break_points = self._get_breakpoints_virtualtime(self.interval, reader.num_of_line)
            if save:
                with open('temp/break_points_' + mode + str(self.interval) + '.dat', 'wb') as ifile:
                    pickle.dump(break_points, ifile)

        return reuse_dist, break_points

        # if reuse_dist_python_list:
        #     return reuse_dist_python_list, break_points
        # else:
        #     return c_reuse_dist_long_array, break_points

    def _prepare_multiprocess_params_LRU(self, plot_type, break_points, **kwargs):
        # 1. hit_rate_start_time_end_time
        # 2. hit_rate_start_time_cache_size
        # 3. avg_rd_start_time_end_time
        # 4. cold_miss_count_start_time_end_time
        # 5. rd_distribution

        # create the array for storing results
        # result is a dict: (x, y) -> heat, x, y is the left, lower point of heat square
        kwargs_dict = {}  # the dictionary for passing parameters to subprocess
        if plot_type == "hit_rate_start_time_cache_size":
            max_rd = max(kwargs['reuse_distance_list'])
            kwargs_dict['max_rd'] = max_rd
            if 'bin_size' in kwargs:
                kwargs_dict["bin_size"] = kwargs['bin_size']
            else:
                kwargs_dict['bin_size'] = 1

            result = np.empty((len(break_points), max_rd // kwargs_dict['bin_size'] + 1), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_hit_rate_start_time_cache_size_subprocess
            kwargs_dict['yticks'] = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(kwargs_dict['bin_size'] * x))
            kwargs_dict['xlabel'] = 'time'
            kwargs_dict['ylabel'] = 'cache size'
            kwargs_dict['zlabel'] = 'hit rate'
            plt.gca().yaxis.set_major_formatter(
                ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(kwargs_dict['bin_size'] * x)))

        elif plot_type == "rd_distribution":
            max_rd = max(kwargs['reuse_distance_list'])
            # print("max rd = %d"%max_rd)
            kwargs_dict['max_rd'] = max_rd
            kwargs_dict['log_num'] = self._cal_log_num(max_rd, len(break_points))
            plt.xlabel("time")
            plt.ylabel('reuse distance')
            plt.gca().yaxis.set_major_formatter(
                ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(kwargs_dict['log_num'] ** x)))

            result = np.empty((len(break_points), int(math.log(max_rd, kwargs_dict['log_num'])) + 1), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_rd_distribution_subprocess_deprecated

        elif plot_type == "hit_rate_start_time_end_time":
            # (x,y) means from time x to time y
            reader = kwargs['reader']
            kwargs_dict['cache_size'] = self.cache_size
            array_len = len(break_points)
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_hit_rate_start_time_end_time_subprocess

            kwargs_dict['yticks'] = ticker.FuncFormatter(
                lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))

            last_access = c_heatmap.get_last_access_dist(reader.cReader)
            last_access_array = Array('l', len(last_access), lock=False)
            for i, j in enumerate(last_access):
                last_access_array[i] = j
            kwargs_dict['last_access_array'] = last_access_array


        elif plot_type == "avg_rd_start_time_end_time":
            # (x,y) means from time x to time y
            kwargs_dict['cache_size'] = self.cache_size

            array_len = len(break_points)
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_avg_rd_start_time_end_time_subprocess

        elif plot_type == "cold_miss_count_start_time_end_time":
            # (x,y) means from time x to time y
            kwargs_dict['cache_size'] = self.cache_size
            array_len = len(break_points)
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan
            func_pointer = calc_cold_miss_count_start_time_end_time_subprocess

        else:
            raise RuntimeError("cannot recognize plot type")

        kwargs_dict['xticks'] = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))

        return result, kwargs_dict, func_pointer

    def calculate_heatmap_dat(self, mode, reader, plot_type, **kwargs):
        """
            calculate the data for plotting heatmap

        :param mode: mode can be either virtual time(v) or real time(r)
        :param reader: for reading the trace file
        :param plot_type: the type of heatmap to generate, possible choices are:
        1. hit_rate_start_time_end_time
        2. hit_rate_start_time_cache_size
        3. avg_rd_start_time_end_time
        4. cold_miss_count_start_time_end_time
        5. rd_distribution
        6.

        :param kwargs: possible key-arg includes the type of cache replacement algorithms(cache),

        :return: a two-dimension list, the first dimension is x, the second dimension is y, the value is the heat value
        """
        if 'calculate' in kwargs:
            calculate = kwargs['calculate']
        else:
            calculate = True

        if 'LRU' not in kwargs or ('LRU' in kwargs and kwargs['LRU']):
            # reuse_dist_python_list, break_points = self._prepare_reuse_distance_and_break_points(mode, reader, calculate)
            reuse_dist, break_points = self._prepare_reuse_distance_and_break_points(mode, reader, calculate)

            # create shared memory for child process
            reuse_dist_share_array = Array('l', len(reuse_dist), lock=False)
            for i, j in enumerate(reuse_dist):
                reuse_dist_share_array[i] = j
            kwargs['reuse_distance_list'] = reuse_dist_share_array

            kwargs['reader'] = reader
            result, kwargs_dict, func_pointer = self._prepare_multiprocess_params_LRU(plot_type, break_points,
                                                                                      **kwargs)  # reuse_distance_list=reuse_dist_share_array,

        else:
            # if it is not LRU, then there should be a cache replacement algorithm in kwargs
            assert 'cache' in kwargs, "you didn't provide any cache replacement algorithm"
            cache_type = kwargs['cache']
            cache_size = kwargs['cache_size']

            # for general cache replacement algorithm, uncomment the following, it is just for
            # transform the data from other format to plain txt
            # p = pardaProfiler(30000, reader)
            p = LRUProfiler(cache_size, reader)

            # prepare break points
            if mode == 'r':
                break_points = self._get_breakpoints_realtime(reader, self.interval)
            elif mode == 'v':
                break_points = self._get_breakpoints_virtualtime(self.interval, p.num_of_lines)
            else:
                raise RuntimeError("unrecognized mode, it can only be r or v")

            kwargs_dict = {}
            array_len = len(break_points)
            result = np.empty((array_len, array_len), dtype=np.float32)
            result[:] = np.nan

        # print(break_points)

        # create shared memory for break points
        break_points_share_array = Array('i', len(break_points), lock=False)
        for i, j in enumerate(break_points):
            break_points_share_array[i] = j

        # Jason: efficiency can be further improved by porting into Cython and improve parallel logic
        # Jason: memory usage can be optimized by not copying whole reuse distance array in each sub process
        map_list = deque()
        for i in range(len(break_points) - 1):
            map_list.append(i)

        # new 0510
        q = Queue()
        process_pool = []
        process_count = 0
        result_count = 0
        map_list_pos = 0
        while result_count < len(map_list):
            if process_count < self.num_of_process and map_list_pos < len(map_list):
                # LRU related heatmaps
                if 'LRU' not in kwargs or ('LRU' in kwargs and kwargs['LRU']):
                    # p = Process(target=_calc_hit_rate_subprocess, args=(map_list[map_list_pos], self.cache_size,
                    #                                                     break_points_share_array, reuse_dist_share_array,
                    #                                                     q))
                    # print(kwargs_dict)
                    p = Process(target=func_pointer, args=(map_list[map_list_pos], break_points_share_array,
                                                           reuse_dist, q), kwargs=kwargs_dict)
                else:
                    p = Process(target=calc_hit_rate_start_time_end_time_subprocess_general,
                                args=(map_list[map_list_pos], cache_type,
                                      break_points_share_array, reader, q), kwargs={'cache_size': cache_size})

                p.start()
                process_pool.append(p)
                process_count += 1
                map_list_pos += 1
            else:
                rl = q.get()
                for r in rl:
                    result[r[0]][r[1]] = r[2]
                process_count -= 1
                result_count += 1
            print("%2.2f%%" % (result_count / len(map_list) * 100), end='\r')
        for p in process_pool:
            p.join()

        # old 0510
        # with Pool(processes=self.num_of_process) as p:
        #     for ret_list in p.imap_unordered(_calc_hit_rate_subprocess, map_list,
        #                                      chunksize=10):
        #         count += 1
        #         print("%2.2f%%" % (count / len(map_list) * 100), end='\r')
        #         for r in ret_list:  # l is a list of (x, y, hr)
        #             result[r[0]][r[1]] = r[2]

        # print(result)
        with open('temp/draw', 'wb') as ofile:
            pickle.dump(result, ofile)

        return result, kwargs_dict

    def _cal_log_num(self, max_rd, length):
        """
        find out the LOG_NUM that makes the number of pixels in y axis to match the number of pixels in x axis
        :param max_rd: maxinum reuse distance
        :param length: length of break points (also number of pixels along x axis)
        :return:
        """
        init_log_num = 10
        l2_prev = length
        while True:
            l2 = math.log(max_rd, init_log_num)
            if l2 > length > l2_prev:
                return init_log_num
            l2_prev = l2
            init_log_num = (init_log_num - 1) / 2 + 1

    @staticmethod
    def _get_breakpoints_virtualtime(bin_size, total_length):
        """

        :param bin_size: the size of the group (virtual time slice)
        :param total_length: the total length of trace requests
        :return: a list of break points
        """
        break_points = []
        assert total_length // bin_size > 10, "bin size too large, please choose a smaller one, " \
                                              "total length %d, bin size %d" % (total_length, bin_size)
        for i in range(total_length // bin_size):
            break_points.append(i * bin_size)
        if break_points[-1] != total_length - 1:
            break_points.append(total_length - 1)

        if (len(break_points)) > 10000:
            colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
            exact size: %d, it may take a very long time, if you didn't intend to do it, \
            please try with a larger time stamp" % len(break_points))

        return break_points

    @staticmethod
    def _get_breakpoints_realtime(reader, interval):
        """
        given reader(vscsi) and time interval, split the data according to the time interval,
        save the order into break_points(list)
        :param reader:
        :param interval:
        :return:
        """
        reader.reset()
        break_points = []
        prev = 0
        for num, line in enumerate(reader.lines()):
            if (float(line[0]) - prev) > interval:
                break_points.append(num)
                prev = float(line[0])

        # noinspection PyUnboundLocalVariable
        if line[0] != prev:
            # noinspection PyUnboundLocalVariable
            break_points.append(num)

        if (len(break_points)) > 10000:
            colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
            exact size: %d, it may take a very long time, if you didn't intend to do it, \
            please try with a larger time stamp" % len(break_points))

        # print(break_points)
        return break_points

    @staticmethod
    def _calc_hit_rate(reuse_dist_array, last_access_dist_array, cache_size, begin_pos, end_pos):
        """
        the original function for calculating hit rate, given the reuse_dist_array, cache_size,
        begin position and end position in the reuse_dist_array, return hit rate
        now isolate as a separate function
        :param reuse_dist_array:
        :param cache_size:
        :param begin_pos:
        :param end_pos:
        :return:
        """
        hit_count = 0
        miss_count = 0
        for i in range(begin_pos, end_pos):
            if reuse_dist_array[i] == -1:
                # never appear
                miss_count += 1
                continue
            if last_access_dist_array[i] - (i - begin_pos) <= 0 and reuse_dist_array[i] < cache_size:
                # hit
                hit_count += 1
                print("{}: {}".format(i, reuse_dist_array[i]))
            else:
                # miss
                miss_count += 1
        print("hit={}, hit+miss={}, total size:{}, hit rate:{}".format(hit_count, hit_count + miss_count,
                                                                       end_pos - begin_pos,
                                                                       hit_count / (end_pos - begin_pos)))
        return hit_count / (end_pos - begin_pos)

    @staticmethod
    def draw_heatmap(xydict, plot_type, **kwargs):
        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = 'heatmap.png'

        masked_array = np.ma.array(xydict, mask=np.isnan(xydict))

        # print(masked_array)
        cmap = plt.cm.jet
        cmap.set_bad('w', 1.)

        try:
            # ax.pcolor(masked_array.T, cmap=cmap)
            if plot_type == "rd_distribution":  # or plot_type == "cold_miss_count_start_time_end_time"
                img = plt.imshow(masked_array.T, interpolation='nearest', origin='lower', aspect='auto'
                                 , norm=matplotlib.colors.LogNorm())
            else:
                if 'fixed_range' in kwargs and kwargs['fixed_range']:
                    img = plt.imshow(masked_array.T, vmin=0, vmax=1, interpolation='nearest', origin='lower',
                                     aspect='auto')
                else:
                    img = plt.imshow(masked_array.T, interpolation='nearest', origin='lower', aspect='auto')

            cb = plt.colorbar(img)
            if 'text' in kwargs:
                (length1, length2) = masked_array.shape
                ax = plt.gca()
                ax.text(length2 // 3, length1 // 8, kwargs['text'], fontsize=20)  # , color='blue')

            if 'xlabel' in kwargs:
                plt.xlabel(kwargs['xlabel'])
            if 'ylabel' in kwargs:
                plt.ylabel(kwargs['ylabel'])
            if 'xticks' in kwargs:
                plt.gca().xaxis.set_major_formatter(kwargs['xticks'])
            if 'yticks' in kwargs:
                plt.gca().yaxis.set_major_formatter(kwargs['yticks'])

            # # change tick from arbitrary number to real time
            if 'change_label' in kwargs and kwargs['change_label'] and 'interval' in kwargs:
                ticks = ticker.FuncFormatter(lambda x, pos: '{:2.2f}'.format(x * kwargs['interval'] / (10 ** 6) / 3600))
                plt.gca().xaxis.set_major_formatter(ticks)
                plt.gca().yaxis.set_major_formatter(ticks)
                plt.xlabel("time/hour")
                plt.ylabel("time/hour")

            # plt.show()
            plt.savefig(filename, dpi=600)
            colorfulPrint("red", "plot is saved at the same directory")
            plt.clf()
        except Exception as e:
            logging.warning(str(e))
            plt.pcolormesh(masked_array.T, cmap=cmap)
            plt.savefig(filename)

    @staticmethod
    def draw2d(l, **kwargs):
        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = '2d_plot.png'

        try:
            # print(l)
            plt.plot(l)

            if 'xlabel' in kwargs:
                plt.xlabel(kwargs['xlabel'])
            if 'ylabel' in kwargs:
                plt.ylabel(kwargs['ylabel'])
            if 'xticks' in kwargs:
                plt.gca().xaxis.set_major_formatter(kwargs['xticks'])
            if 'yticks' in kwargs:
                plt.gca().yaxis.set_major_formatter(kwargs['yticks'])

            # plt.show()
            plt.savefig(filename)
            colorfulPrint("red", "plot is saved at the same directory")
            plt.clf()
        except Exception as e:
            logging.warning(str(e))
            plt.savefig(filename)

    @staticmethod
    def draw_3D(xydict, **kwargs):
        from mpl_toolkits.mplot3d import Axes3D

        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = 'heatmap_3D.png'

        fig = plt.figure()
        ax = Axes3D(fig)
        # X = np.arange(-4, 4, 0.25)
        X = np.arange(0, len(xydict))
        # Y = np.arange(-4, 4, 0.25)
        Y = np.arange(0, len(xydict[0]))
        X, Y = np.meshgrid(X, Y)
        # R = np.sqrt(X ** 2 + Y ** 2)
        # Z = np.sin(R)
        Z = np.array(xydict).T

        print(X.shape)
        print(Y.shape)
        print(Z.shape)

        # print(X)
        # print(Y)
        # print(Z)

        if 'xlabel' in kwargs:
            ax.set_xlabel(kwargs['xlabel'])
        if 'ylabel' in kwargs:
            ax.set_ylabel(kwargs['ylabel'])
        if 'xticks' in kwargs:
            print("setx")
            ax.xaxis.set_major_formatter(kwargs['xticks'])
        if 'yticks' in kwargs:
            print("sety")
            ax.yaxis.set_major_formatter(kwargs['yticks'])

        ax.plot_surface(X[:, :-3], Y[:, :-3], Z[:, :-3], rstride=10000, cstride=2, cmap=plt.cm.jet)  # plt.cm.hot
        ax.contourf(X, Y, Z, zdir='z', offset=0, cmap=plt.cm.jet)
        ax.set_zlim(0, 1)
        # ax.set_zlim(-2, 2)

        plt.show()
        plt.savefig(filename, dpi=600)

    def __del_manual__(self):
        """
        cleaning
        :return:
        """
        if os.path.exists('temp/'):
            for filename in os.listdir('temp/'):
                os.remove('temp/' + filename)
            os.rmdir('temp/')

    def run(self, mode, interval, plot_type, reader, **kwargs):
        """

        :param plot_type:
        :param mode:
        :param interval:
        :param reader:
        :param kwargs: include num_of_process, figname
        :return:
        """
        self.cache_size = kwargs['cache_size']
        self.interval = interval
        self.mode = mode
        reader.reset()
        if 'num_of_process' in kwargs:
            self.num_of_process = kwargs['num_of_process']
        else:
            self.num_of_process = 4

        if mode == 'r' or mode == 'v':
            xydict, kwargs_dict = self.calculate_heatmap_dat(mode, reader, plot_type, interval=interval, **kwargs)
        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)

        kwargs_dict.update(kwargs)
        self.draw_heatmap(xydict, plot_type, interval=interval, **kwargs_dict)
        reader.reset()
        # self.draw_test(xydict, **kwargs_dict)

        # self.__del_manual__()

    def draw_diff(self, xydict1, xydict2, plot_type, **kwargs):

        # print(xydict1.shape)
        # print(xydict2.shape)
        print(xydict1)
        print(xydict2)

        xydict = xydict1 - xydict2

        for i in range(len(xydict)):
            for j in range(len(xydict[i])):
                if xydict[i][j] < 0:
                    print("({}, {})".format(i, j))

        self.draw_heatmap(xydict, plot_type, **kwargs)

    def run_diff_optimal(self, mode, interval, plot_type, reader, **kwargs):
        """

        :param plot_type:
        :param mode:
        :param interval:
        :param reader:
        :param kwargs: include num_of_process, figname
        :return:
        """
        self.cache_size = kwargs['cache_size']
        self.interval = interval
        self.mode = mode
        if 'num_of_process' in kwargs:
            self.num_of_process = kwargs['num_of_process']
        else:
            self.num_of_process = 4

        if mode == 'r':
            xydict1, kwargs_dict1 = self.calculate_heatmap_dat('r', reader, plot_type, LRU=False, cache='optimal',
                                                               num_of_process=self.num_of_process,
                                                               cache_size=kwargs['cache_size'])

            reader.reset()
            xydict2, kwargs_dict2 = self.calculate_heatmap_dat('r', reader, plot_type, **kwargs)

        elif mode == 'v':
            xydict1, kwargs_dict1 = self.calculate_heatmap_dat('v', reader, plot_type, LRU=False, cache='optimal',
                                                               num_of_process=self.num_of_process,
                                                               cache_size=kwargs['cache_size'])
            reader.reset()
            xydict2, kwargs_dict2 = self.calculate_heatmap_dat('v', reader, plot_type, **kwargs)
        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)
        kwargs_dict = {}
        kwargs_dict.update(kwargs_dict1)
        kwargs_dict.update(kwargs_dict2)
        kwargs_dict.update(kwargs)
        self.draw_diff(xydict1, xydict2, plot_type, **kwargs_dict)

        # self.__del_manual__()


def server_plot_all():
    mem_sizes = []
    with open('memSize', 'r') as ifile:
        for line in ifile:
            mem_sizes.append(int(line.strip()))

    for filename in os.listdir("../data/cloudphysics"):
        print(filename)
        if filename.endswith('.vscsitrace'):
            # if int(filename.split('_')[0][1:]) in [1, 3, 4, 5, 51, 99, 83, 87]:
            #     continue

            hm = heatmap()
            mem_size = mem_sizes[int(filename.split('_')[0][1:]) - 1] * 16
            reader = vscsiCacheReader("../data/cloudphysics/" + filename)
            if os.path.exists("0601/" + filename + "LRU_Optimal_" + str(mem_size) + "_r.png"):
                continue

            # cold_miss_2d(reader, 'r', 10000000, figname='0601/' + filename + '_cold_miss_r.png')
            # request_num_2d(reader, 'r', 10000000, figname='0601/' + filename + '_request_num_r.png')
            #
            # hm.run('r', 1000000000, "hit_rate_start_time_cache_size", reader, num_of_process=48, cache_size=mem_size,
            #        save=True, bin_size=1000, text="size = " + str(mem_size), fixed_range=True,
            #        figname='0601/' + filename + '_hit_rate_start_time_cache_size_r.png')
            # # ,change_label='False'     ,   10000000
            #
            #
            # hm.run('r', 1000000000, "avg_rd_start_time_end_time", reader, num_of_process=48,
            #        # LRU=False, cache='optimal',
            #        calculated=True, cache_size=mem_size,
            #        figname='0601/' + filename + "_avg_rd_start_time_end_time_r.png")
            #
            # hm.run('r', 1000000000, "rd_distribution", reader, num_of_process=48,  # LRU=False, cache='optimal',
            #        calculated=True, cache_size=mem_size,
            #        figname='0601/' + filename + "_rd_distribution_r.png")
            #
            # hm.run('r', 1000000000, "cold_miss_count_start_time_end_time", reader, num_of_process=48,
            #        calculated=True, cache_size=mem_size,
            #        figname='0601/' + filename + "_cold_miss_count_start_time_end_time_r.png")
            #
            # hm.run('r', 1000000000, "hit_rate_start_time_end_time", reader, num_of_process=48,
            #        cache_size=mem_size, calculated=True, fixed_range=True, change_label=True,
            #        figname='0601/' + filename + "_hit_rate_start_time_end_time_" + str(mem_size) + "_r.png")
            # reader.reset()


            hm.run_diff_optimal('r', 1000000000, "hit_rate_start_time_end_time", reader, num_of_process=48,
                                cache_size=mem_size, save=False, change_label=True,
                                figname='0601/' + filename + "LRU_Optimal_" + str(mem_size) + "_r.png")

            del hm
            del reader


def server_size_plot():
    mem_sizes = []
    with open('memSize', 'r') as ifile:
        for line in ifile:
            mem_sizes.append(int(line.strip()))

    for filename in os.listdir("../data/cloudphysics"):
        if filename.endswith('.vscsitrace'):
            if int(filename.split('_')[0][1:]) in [1, 4, 5, 51, 99, 83, 87]:
                continue
            if os.path.exists('time_size_0523/' + filename + '_128.png'):
                continue
            hm = heatmap()
            mem_size = mem_sizes[int(filename.split('_')[0][1:]) - 1]
            reader = vscsiCacheReader("../data/cloudphysics/" + filename)
            print(filename)
            size = 128
            while size < mem_size * 1024:
                print(size)
                if size == 128:
                    hm.run('r', 10000000, "hit_rate_start_time_end_time", reader, num_of_process=48,
                           figname='time_size/' + filename + '_' + str(size) + '.png',
                           cache_size=size, change_label=True, save=True,
                           calculate=True, fixed_range=True, text='size=' + str(size))
                else:
                    hm.run('r', 10000000, "hit_rate_start_time_end_time", reader, num_of_process=48,
                           figname='time_size/' + filename + '_' + str(size) + '.png',
                           cache_size=size, change_label=True, save=True,
                           calculate=False, fixed_range=True, text='size=' + str(size))
                size *= 2


def server_plot_all_redis():
    for filename in os.listdir("../data/redis/"):
        print(filename)
        if filename.endswith('.csv'):
            # print(reader)
            # for line in reader:
            #     print(line)
            #     break

            if filename in []:
                continue
            if os.path.exists(filename + '_r.png'):
                continue
            hm = heatmap()
            reader = csvCacheReader("../data/redis/" + filename, column=1)
            hm.run('r', 20, 2000, reader, num_of_process=48, figname=filename + '_r.png',
                   fixed_range="True")  # , change_label='True')


def request_num_2d(reader, mode, interval, **kwargs):
    # reader = vscsiCacheReader("../data/trace_CloudPhysics_bin")

    hm = heatmap()
    if mode == 'r':
        break_points = hm._get_breakpoints_realtime(reader, interval)
    elif mode == 'v':
        # p = pardaProfiler(2000, reader2)
        p = LRUProfiler(2000, reader)
        break_points = hm._get_breakpoints_virtualtime(interval, p.num_of_lines)

    l = []
    for i in range(len(break_points)):
        if i == 0:
            continue
        else:
            l.append(break_points[i] - break_points[i - 1])
    xticks = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))
    hm.draw2d(l, figname=kwargs['figname'], xticks=xticks)

    reader.reset()


def cold_miss_2d(reader, mode, interval, **kwargs):
    # reader2 = vscsiCacheReader("../data/trace_CloudPhysics_bin")

    hm = heatmap()
    break_points = None
    if mode == 'r':
        break_points = hm._get_breakpoints_realtime(reader, interval)
    elif mode == 'v':
        p = pardaProfiler(2000, reader)
        break_points = hm._get_breakpoints_virtualtime(interval,
                                                       reader.get_num_of_total_requests())  # p.num_of_trace_elements)

    l = calc_cold_miss_count(break_points, reader)
    xticks = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))
    hm.draw2d(l, figname=kwargs['figname'], xticks=xticks)
    reader.reset()


def localtest():
    CACHE_SIZE = 2000
    # reader1 = plainCacheReader("../data/parda.trace")



    reader2 = vscsiCacheReader("../data/trace_CloudPhysics_bin")

    hm = heatmap()

    # cold_miss_2d('r', 10000000, filename='cold_miss.png')
    # request_num_2d('r', 10000000, filename='request_num.png')

    # hm.run('r', 100000000, "hit_rate_start_time_cache_size", reader2, num_of_process=8,  # LRU=False, cache='optimal',
    #        save=False, fixed_range=True, cache_size=CACHE_SIZE, bin_size=200,
    #        figname="0529/3D_small_LRU_hit_rate_start_time_cache_size.png")
    # reader2.reset()

    # hm.run('r', 10000000, "avg_rd_start_time_end_time", reader2, num_of_process=8,  # LRU=False, cache='optimal',
    #        save=False, cache_size=CACHE_SIZE,
    #        figname="0529/small_LRU_avg_rd_start_time_end_time.png")
    # reader2.reset()

    hm.run('r', 10000000, "rd_distribution", reader2, num_of_process=8,  # LRU=False, cache='optimal',
           save=False, cache_size=CACHE_SIZE,
           figname="small_LRU_rd_distribution1.png")
    reader2.reset()

    # hm.run('r', 10000000, "cold_miss_count_start_time_end_time", reader2, num_of_process=8,
    #        # LRU=False, cache='optimal',
    #        save=False, cache_size=CACHE_SIZE,
    #        figname="0529/small_LRU_cold_miss_count_start_time_end_time.png")
    # reader2.reset()


    # hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8,  # LRU=False, cache='optimal',
    #        cache_size=CACHE_SIZE, save=False, fixed_range=True, change_label=True,
    #        figname="small_LRU_hit_rate_start_time_end_time_" + str(CACHE_SIZE) + ".png")


    reader2.reset()

    # hm.run_diff_optimal('r', 100000000, "hit_rate_start_time_end_time", reader2, num_of_process=47,
    #                     # LRU=False, cache="LRU",
    #                     cache_size=CACHE_SIZE, save=False, change_label=True,
    #                     figname="LRU_NONPARDA_Optimal_" + str(CACHE_SIZE) + "3.png")

    # reader2.reset()


'''
    for CACHE_SIZE in range(30, 1200, 120):
        hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8, # LRU=False, cache='optimal',
               cache_size=CACHE_SIZE, save=False, fixed_range=True, change_label=True,
               figname="0529/small_LRU_hit_rate_start_time_end_time_" + str(CACHE_SIZE) + ".png")
        reader2.reset()




        hm.run_diff_optimal('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8, LRU=False, cache="LRU",
                            cache_size=CACHE_SIZE, save=False, change_label=True,
                            figname="0529/LRU_NONPARDA_Optimal_" + str(CACHE_SIZE) + ".png")
        reader2.reset()
        hm.run_diff_optimal('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8, LRU=False,
                            cache="ARC",
                            cache_size=CACHE_SIZE, save=False, change_label=True,
                            figname="0529/ARC_NONPARDA_Optimal_" + str(CACHE_SIZE) + ".png")
        reader2.reset()

        hm.run_diff_optimal('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8, LRU=False,
                            cache="RR",
                            cache_size=CACHE_SIZE, save=False, change_label=True,
                            figname="0529/RR_NONPARDA_Optimal_" + str(CACHE_SIZE) + ".png")
        reader2.reset()

        hm.run_diff_optimal('r', 10000000, "hit_rate_start_time_end_time", reader2, num_of_process=8, LRU=False,
                            cache="SLRU",
                            cache_size=CACHE_SIZE, save=False, change_label=True,
                            figname="0529/SLRU_NONPARDA_Optimal_" + str(CACHE_SIZE) + ".png")
        reader2.reset()



    # hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, LRU=False, num_of_process=8,
    #        cache='LFU_RR', cache_size=500, save=False, change_label=True, figname="LFU_RR.png")
    # reader2.reset()
    # hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, LRU=False, num_of_process=8,
    #        cache='ARC', cache_size=500, save=False, change_label=True, figname="ARC.png")
    # reader2.reset()
    # hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, LRU=False, num_of_process=8,
    #        cache='optimal', cache_size=500, save=False, change_label=True, figname="optimal.png")
    # hm.run('r', 10000000, "hit_rate_start_time_end_time", reader2, LRU=True, num_of_process=8,
    #       cache_size=500, save=False, change_label=True, figname="LRU.png")



    # p = pardaProfiler(20000, reader2)
    # p.run()
'''

if __name__ == "__main__":
    import time

    t1 = time.time()
    localtest()
    # server_plot_all()
    # cold_miss_2d('v', 100)
    t2 = time.time()
    print(t2 - t1)

    # server_plot_all()
    # server_plot_all_redis()
    # server_size_plot()
    # server_request_num_plot()



# from distutils.core import setup, Extension
from setuptools import find_packages, setup, Extension
from glob import glob
from distutils.command.build_ext import build_ext
from distutils.command.build_clib import build_clib
import distutils.sysconfig
import platform
import os
import sys
import shutil
import subprocess
import tempfile
# from Cython.Build import cythonize
from _version import __version__

# --------------------- Initialization ------------------------------

extensions = []
extra_compile_args = []
extra_link_args = ["-lm"]
numpy_headers = []

# --------------------- Get OpenMP compiler flag --------------------

# If this C test program compiles the compiler supports OpenMP
# http://stackoverflow.com/questions/16549893/programatically-testing-for-openmp-support-from-a-python-setup-script
omp_test = \
    r"""
    #include <omp.h>
    #include <stdio.h>
    int main()
    {
        #pragma omp parallel
        printf("Hello from thread %d, nthreads %d\n", omp_get_thread_num(), omp_get_num_threads());
        return 0;
    }
    """


def get_compiler_openmp_flag():
    openmp_flag = ''
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    filename = r'omp_test.c'

    with open(filename, 'w') as file:
        file.write(omp_test)
        file.flush()

    try:
        cc = os.environ['CC']
    except KeyError:
        cc = 'cc'

    # Compile omp_test.c program using different OpenMP compiler flags.
    # If the code below fails continue without OpenMP support.
    with open(os.devnull, 'w') as fnull:
        exit_code = 1
        try:
            exit_code = subprocess.call([cc, '-fopenmp', filename], stdout=fnull, stderr=fnull)
        except:
            pass
        if exit_code == 0:
            openmp_flag = '-fopenmp'
        else:
            try:
                exit_code = subprocess.call([cc, '-openmp', filename], stdout=fnull, stderr=fnull)
            except:
                pass
            if exit_code == 0:
                openmp_flag = '-openmp'

    # clean up
    os.chdir(curdir)
    shutil.rmtree(tmpdir)
    return openmp_flag


def getGlibFlag():
    try:
        glib_cflag = subprocess.check_output("pkg-config --cflags glib-2.0", shell=True).decode().strip()
        extra_compile_args.extend(glib_cflag.split())
    except Exception as e:
        print(e)


def getGlibLibrary():
    try:
        glib_lib = subprocess.check_output("pkg-config --libs glib-2.0 --libs gthread-2.0", shell=True).decode().strip()
        extra_link_args.extend(glib_lib.split())
    except Exception as e:
        print(e)


def setPlatformRelatedConfig():
    if sys.platform == 'darwin':
        from distutils import sysconfig
        vars = sysconfig.get_config_vars()
        vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-dynamiclib')
        extra_link_args.append('-dynamiclib')
        # extra_compile_args.append('-dynamiclib')
        os.environ["CC"] = "clang"
        os.environ["CXX"] = "clang"


def getNumpyHeader():
    try:
        import numpy
        numpy_headers.append(numpy.get_include())
    except Exception as e:
        print(e)


def remove_inline_kw(path='mimircache/CExtension/cache'):
    for file in os.listdir(path):
        if file.endswith('.c'):
            shutil.copyfile(path + '/' + file, path + '/' + file + '.temp')



getGlibFlag()
getGlibLibrary()
getNumpyHeader()
setPlatformRelatedConfig()
# openmp_flag = get_compiler_openmp_flag()
# print('get_compiler_openmp_flag(): ' + openmp_flag)

# extra_compile_args.append(openmp_flag)
# extra_link_args.append(openmp_flag)


print('all compile flags: ' + str(extra_compile_args))
print('all link flasgs: ' + str(extra_link_args))

# --------------------- parda module ---------------------------
# extensions.append(Extension(
#     "mimircache.profiler.libparda",
#     glob("mimircache/profiler/parda/src/*.c"),
#     include_dirs=["mimircache/profiler/parda/src/h/"],
#     extra_compile_args=extra_compile_args,
#     extra_link_args=extra_link_args,
#     language="c"
# ))

# # --------------------- vscsi module ---------------------------
extensions.append(Extension(
    "mimircache.cacheReader.libvscsi",
    glob("mimircache/cacheReader/vscsi/src/*.c"),
    include_dirs=["mimircache/cacheReader/vscsi/src/"],
    language="c"
))

extensions.append(Extension(
    'mimircache.c_cacheReader',
    glob("mimircache/CExtension/cacheReader/*.c"),
    include_dirs=["mimircache/CExtension/headers"] + numpy_headers,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c"))

extensions.append(Extension(
    'mimircache.c_LRUProfiler',
    glob("mimircache/CExtension/profiler/LRUProfiler/*.c") +
    glob('mimircache/CExtension/cacheReader/*.c') +
    glob('mimircache/CExtension/utils/*.c'),
    include_dirs=["mimircache/CExtension/headers"] + numpy_headers, 
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c"))

extensions.append(Extension(
    'mimircache.c_generalProfiler',
    glob("mimircache/CExtension/profiler/*.c") +
    glob("mimircache/CExtension/profiler/generalProfiler/*.c") +
    glob("mimircache/CExtension/profiler/LRUProfiler/*.c") +
    glob("mimircache/CExtension/cache/*.c") +
    glob('mimircache/CExtension/cacheReader/*.c') +
    glob('mimircache/CExtension/utils/*.c') +
    glob('mimircache/CExtension/wrapper/*.c'),
    include_dirs=["mimircache/CExtension/headers/"] +
                 ["mimircache/CExtension/headers/cache/"] + numpy_headers,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c"))

extensions.append(Extension(
    'mimircache.c_heatmap',
    glob("mimircache/CExtension/profiler/*.c") +
    glob("mimircache/CExtension/profiler/generalProfiler/*.c") +
    glob("mimircache/CExtension/profiler/LRUProfiler/*.c") +
    glob("mimircache/CExtension/cache/*.c") +
    glob('mimircache/CExtension/cacheReader/*.c') +
    glob('mimircache/CExtension/utils/*.c') +
    glob('mimircache/CExtension/wrapper/*.c'),
    include_dirs=["mimircache/CExtension/headers"] +
                 ["mimircache/CExtension/headers/cache"] + numpy_headers,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    language="c"))

# --------------------- parda module in c_lib---------------------------
# parda_include = []
# for i in extra_compile_args:
#     if i.startswith('-I'):
#         parda_include.append(i[2:])
# parda_include.append("mimircache/profiler/parda/src/h/")
# print(parda_include)

# libparda = ("mimircache.profiler.libparda", dict(
#     sources = glob("mimircache/profiler/parda/src/*.c"),
#     include_dirs = parda_include,
#     extra_compile_args = extra_compile_args,
#     # extra_link_args = extra_link_args,
#     language = "c"
#     ))

# --------------------- vscsi module in c_lib---------------------------
# libvscsi = ("mimircache.cacheReader.libvscsi", dict(
#     sources = glob("mimircache/cacheReader/vscsi/src/*.c"),
#     include_dirs = ["mimircache/cacheReader/vscsi/src/"],
#     language = "c"
#     ))


# extensions.append(cythonize("mimircache/cache/RR.pyx"))



print("find packages: " + str(
    find_packages(exclude=(['mimircache.bin', 'mimircache.test', 'mimircache.data', 'mimircache.1a1a11a']))))

print("Extension: " + str(extensions))
# long_description = read('README.md', 'CHANGES')


setup(
    name="mimircache",
    version=__version__,
    # package_dir = {'':'src'},
    # packages = ['cache', 'cacheReader', 'profiler', 'utils'],
    packages=find_packages(exclude=(['mimircache.bin', 'mimircache.test', 'mimircache.data', 'mimircache.1a1a11a'])),
    # modules = 
    package_data={'plain': ['mimircache/data/trace.txt'],
                  'csv': ['mimircache/data/trace.csv'],
                  'vscsi': ['mimircache/data/trace.vscsi'],
                  'conf': ['mimircache/conf']},
    include_package_data=True,
    author="Juncheng Yang",
    author_email="peter.waynechina@gmail.com",
    description="mimircache platform for analyzing cache traces, developed by Ymir group @ Emory University",
    license="GPLv3",
    keywords="mimircache cache Ymir",
    url="http://www.google.com",

    # libraries = [libparda, libvscsi],
    # cmdclass = {'build_clib' : build_clib},
    # cmdclass = {'build_ext': build_ext_subclass, 'build_clib' : build_clib_subclass},

    ext_modules=extensions,
    classifiers=[
        'Development Status :: 3 - Alpha',

        # 'Topic :: Operating System :: cache analysis'

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    # install_requires=['numpy', 'matplotlib', 'scipy'],
    # long_description=long_description
)



# python3 setup.py sdist upload -r pypitest

# could also try this: 
# from distutils.ccompiler import new_compiler

# # Create compiler with default options
# c = new_compiler()
# workdir = "."

# # Optionally add include directories etc.
# # c.add_include_dir("some/include")

# # Compile into .o files
# objects = c.compile(["a.c", "b.c"])

# # Create static or shared library
# c.create_static_lib(objects, "foo", output_dir=workdir)
# c.link_shared_lib(objects, "foo", output_dir=workdir)



import glob
import os

import imageio
from PIL import Image

DIRECTORY = 'temp/'


def sort_filenames(filenames):
    return sorted(filenames, key=lambda x: int(x.split('/')[-1].split('_')[-1][:-4]))


def offline_plotting(folder_loc):
    if not os.path.exists(DIRECTORY):
        os.makedirs(DIRECTORY)
    gif_list = []
    for fileloc in glob.glob(folder_loc + "/*.png"):
        if not fileloc.split('/')[-1].split('_')[0] in gif_list:
            gif_list.append(fileloc.split('/')[-1].split('_')[0])
        im = Image.open(fileloc)
        im.save(DIRECTORY + fileloc.split('/')[-1])

    for prefix in gif_list:
        images = []
        png_list = []
        for fileloc in glob.glob(DIRECTORY + prefix + '*.png'):
            png_list.append(fileloc)
        print(png_list)
        png_list = sort_filenames(png_list)
        print(png_list)
        for fileloc in png_list:
            images.append(imageio.imread(fileloc))
        imageio.mimsave(prefix + '.gif', images, duration=0.6)


offline_plotting('time_size')



import string
from mimircache.cacheReader.abstractReader import cacheReaderAbstract


class csvCacheReader(cacheReaderAbstract):
    def __init__(self, file_loc, column, header=False, delimiter=','):
        super(csvCacheReader, self).__init__(file_loc)
        self.trace_file = open(file_loc, 'r')

        self.column = column
        self.counter = 0
        self.header_bool = header
        self.delimiter = delimiter
        if header:
            self.headers = [i.strip(string.whitespace) for i in self.trace_file.readline().split(self.delimiter)]

    def read_one_element(self):
        super().read_one_element()
        line = self.trace_file.readline()
        if line:
            return line.split(self.delimiter)[self.column].strip()
        else:
            return None

    def lines_dict(self):
        line = self.trace_file.readline()
        while line:
            line_split = line.split(self.delimiter)
            d = {}
            if self.header_bool:
                for i in range(len(self.headers)):
                    d[self.headers[i]] = line_split[i].strip(string.whitespace)
            else:
                for key, value in enumerate(line_split):
                    d[key] = value
            line = self.trace_file.readline()
            yield d

    def lines(self):
        line = self.trace_file.readline()
        while line:
            line_split = tuple(line.split(self.delimiter))
            line = self.trace_file.readline()
            yield line_split


    def __next__(self):  # Python 3
        super().__next__()
        element = self.trace_file.readline().strip()
        if element:
            return element.split(self.delimiter)[self.column].strip()
        else:
            raise StopIteration

    def __repr__(self):
        return "csv cache reader {}, specified column: {}, column begins from 0".format(self.file_loc, self.column)




if __name__ == "__main__":
    reader = csvCacheReader('../data/trace_CloudPhysics_txt2', 4, header=True, delimiter=',')

    # usage one: for reading all elements
    # for i in reader:
    #     print(i)

    for i in reader.lines_dict():
        print(i['op'])

        # usage two: best for reading one element each time
        # s = reader.read_one_element()
        # # s2 = next(reader)
        # while (s):
        #     print(s)
        #     s = reader.read_one_element()
        # s2 = next(reader)



import os
import configparser
from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader

from mimircache.cache.ARC import ARC
from mimircache.cache.clock import clock
from mimircache.cache.FIFO import FIFO
from mimircache.cache.LFU_LRU__NEED_OPTIMIZATION import LFU_LRU
from mimircache.cache.LFU_MRU import LFU_MRU
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.MRU import MRU
from mimircache.cache.Random import Random
from mimircache.cache.SLRU import SLRU
from mimircache.cache.S4LRU import S4LRU
from mimircache.cache.Optimal import optimal
from mimircache.cache.ARC import ARC

# global c_available_cache
c_available_cache = []
c_available_cacheReader = [plainCacheReader, vscsiCacheReader, csvCacheReader]
cache_alg_mapping = {}
BASE_DIR = os.path.dirname(__file__)

DEFAULT_BIN_NUM_PROFILER = 100
DEFAULT_NUM_OF_THREADS = 4


def init():
    _init_C_available_modules()
    _init_cache_alg_mapping()


def _init_C_available_modules():
    config = configparser.ConfigParser()
    # print(BASE_DIR + '/conf')
    config.read(BASE_DIR + '/conf')
    if 'C_available_cache' in config.sections():
        c_available_cache.extend(config['C_available_cache'])
    else:
        raise RuntimeWarning("cannot find any cache module in C")

        # if 'C_available_reader' in config.sections():
        #     c_available_cacheReader.extend(config['C_available_reader'])
        # else:
        #     raise RuntimeWarning("cannot find any cacheReader module in C")





def _init_cache_alg_mapping():
    """
    match all possible cache replacement algorithm names(lower case) to available cache replacement algorithms
    :return:
    """

    cache_alg_mapping['optimal'] = 'Optimal'
    cache_alg_mapping['rr'] = "Random"
    cache_alg_mapping['random'] = "Random"
    cache_alg_mapping['lru'] = "LRU"
    cache_alg_mapping['fifo'] = "FIFO"
    # cache_alg_mapping['adaptiveSLRU'] = "AdaptiveSLRU"
    cache_alg_mapping['arc'] = "ARC"
    cache_alg_mapping['clock'] = "clock"
    cache_alg_mapping['mru'] = "MRU"
    cache_alg_mapping['slru'] = "SLRU"
    cache_alg_mapping['s4lru'] = "S4LRU"
    cache_alg_mapping['lfu_rr'] = "LFU_RR"
    cache_alg_mapping['lfu_mru'] = "LFU_MRU"
    cache_alg_mapping['lru_k'] = "LRU_K"
    cache_alg_mapping['lru_2'] = "LRU_2"


def cache_name_to_class(name):
    if name.lower() in cache_alg_mapping:
        cache = cache_alg_mapping[name.lower()]
        if cache == 'Random':
            cache_class = Random
        elif cache == 'SLRU':
            cache_class = SLRU
        elif cache == 'S4LRU':
            cache_class = S4LRU
        elif cache == 'ARC':
            cache_class = ARC
        elif cache == 'LFU_RR':
            cache_class = LFU_RR
        elif cache == 'LRU':
            cache_class = LRU
        elif cache == "Optimal":
            cache_class = optimal
        elif cache == 'FIFO':
            cache_class = FIFO
        elif cache == "MRU":
            cache_class = MRU
        elif cache == 'clock':
            cache_class = clock
        elif cache == 'FIFO':
            cache_class = FIFO
        elif cache == 'LFU_MRU':
            cache_class = LFU_MRU
    else:
        raise RuntimeError("cannot recognize given cache replacement algorithm " + str(name))
    return cache_class



"""
.. module:: useful_1
   :platform: Unix, Windows
   :synopsis: A useful module indeed.

.. moduleauthor:: Andrew Carter <andrew@invalid.com>


"""

import abc


class cache:
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __init__(self, cache_size):
        self.cache_size = cache_size
        if self.cache_size <= 0:
            raise RuntimeError("cache size cannot be smaller than or equal 0")

    @abc.abstractmethod
    def checkElement(self, element):
        """
        :param element: the element for search
        :return: whether the given element is in the cache
        """
        raise NotImplementedError("check_element class is not implemented")

    @abc.abstractmethod
    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: True on success, False on failure
        """
        raise NotImplementedError("__update_element__ class is not implemented")

    @abc.abstractmethod
    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        raise NotImplementedError("__insert_element__ class is not implemented")

    @abc.abstractmethod
    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        raise NotImplementedError("__evict_one_element__ class is not implemented")

    @abc.abstractmethod
    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: -1 if not in cache, otherwise old rank (if there is) or 1
        """
        raise NotImplementedError("add_element class is not implemented")

    @abc.abstractmethod
    def _printCacheLine(self):
        return

    def __contains__(self, item):
        return self.checkElement(item)

    def __repr__(self):
        return "abstract cache class, {}".format(super().__repr__())

    def __str__(self):
        return "abstract cache class, {}".format(super().__str__())
